﻿//**********************************//
// Written by: Per Böhlin
// perbo770@student.liu.se
// Copyright: (c) 2009
//
// Implementation started: July 27, 2009
// Last modified: August 12, 2009
//**********************************//

using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using NUnit.Framework;
using BSM.core;

namespace BSM.Tests
{
    /// <summary>
    /// Test fixture for Scheduler
    /// </summary>
    [TestFixture]
    class SchedulerTest
    {
        Scheduler testSubject;
        IJob job;

        [SetUp]
        public void Setup()
        {
            testSubject = new Scheduler();
            job = new Mocks.JobMock(new Status());
            job.Title = "mock";
        }

        [TearDown]
        public void TearDown()
        {
            testSubject.Close();
            testSubject.WaitForAllStartedJobsToFinish();
        }

        /// <summary>
        /// Tests method CreateJob(string title)
        /// </summary>
        [Test]
        public void TestAddJob()
        {
            testSubject.AddJob(job);
            IJob retrievedJob = testSubject.GetJob(job.Title);

            Assert.AreSame(job, retrievedJob);
            Assert.IsTrue(testSubject.GetJobExists(job.Title));
        }



        /// <summary>
        /// Tests method RemoveJob(string title)
        /// </summary>
        [Test]
        public void TestRemoveJob()
        {
            testSubject.AddJob(job);
            testSubject.RemoveJob(job.Title);
            Assert.AreEqual(0, testSubject.JobTitles.Length);
            Assert.IsFalse(testSubject.GetJobExists(job.Title));
        }

        /// <summary>
        /// Tests that a KeyNotFoundException is thrown when
        /// a none existing job is accessed.
        /// </summary>
        [Test]
        public void TestGetNoneExistingJob()
        {
            Assert.Throws(typeof(System.Collections.Generic.KeyNotFoundException), RetrieveNoneExistingJob);

        }

        /// <summary>
        /// Wrapper function. Matches TestDelegate.
        /// Help function for TestGetNoneExistingJob()
        /// </summary>
        public void RetrieveNoneExistingJob()
        {
            IJob retrievJob = testSubject.GetJob("title");
        }

        /// <summary>
        /// Tests IsRunning-property with
        /// Start() and Stop() methods.
        /// </summary>
        [Test]
        public void TestIsRunning()
        {
            Assert.IsFalse(testSubject.IsRunning);
            testSubject.Start();
            Assert.IsTrue(testSubject.IsRunning);
            testSubject.Stop();
            Assert.IsFalse(testSubject.IsRunning);
        }

        /// <summary>
        /// Test that JobTitles property returns
        /// a proper string array with job titles
        /// </summary>
        [Test]
        public void TestGetSortedJobTitles()
        {
            job.Title = "b";
            testSubject.AddJob(job);
            IJob extraJob = new Mocks.JobMock();
            extraJob.Title = "a";
            testSubject.AddJob(extraJob);

            string[] jobTitles = testSubject.JobTitles;
            Assert.AreEqual(2, jobTitles.Length);
            Assert.AreEqual(extraJob.Title, jobTitles[0]);
            Assert.AreEqual(job.Title, jobTitles[1]);
        }

        /// <summary>
        /// Tests the method RunJobs
        /// with valid statuses
        /// </summary>
        [Test]
        public void TestRunJobsWithValidStatus()
        {
            IJob validStatusJob = new Mocks.JobMock(new Status());
            validStatusJob.Title = "valid";

            testSubject.AddJob(validStatusJob);

            Assert.AreEqual(0, ((Mocks.JobMock)validStatusJob).CallsToRun);

            //Method being tested, first run
            testSubject.RunJobs();
            Thread.Sleep(10);
            testSubject.WaitForAllStartedJobsToFinish();

            Assert.AreEqual(1, ((Mocks.JobMock)validStatusJob).CallsToRun);

            //Method being tested, second run
            testSubject.RunJobs();
            Thread.Sleep(10);
            testSubject.WaitForAllStartedJobsToFinish();

            Assert.AreEqual(2, ((Mocks.JobMock)validStatusJob).CallsToRun);
        }

        /// <summary>
        /// Tests the method RunJobs
        /// with null statuses
        /// </summary>
        [Test]
        public void TestRunJobsNullStatus()
        {
            IJob nullStatusJob = new Mocks.JobMock(null);
            nullStatusJob.Title = "null";

            testSubject.AddJob(nullStatusJob);

            Assert.AreEqual(0, ((Mocks.JobMock)nullStatusJob).CallsToRun);

            //Run job which return null statuses.
            testSubject.RunJobs();
            Thread.Sleep(10);
            testSubject.WaitForAllStartedJobsToFinish();
            
            //Run should have been called with no exception thrown
            Assert.AreEqual(1, ((Mocks.JobMock)nullStatusJob).CallsToRun);
        }

        /// <summary>
        /// Tests the method GetJobStatus()
        /// with valid values
        /// </summary>
        [Test]
        public void TestGetJobStatusWithValidStatus()
        {
            Status status = new Status();
            IJob validStatusJob = new Mocks.JobMock(status);
            validStatusJob.Title = "valid";

            //No job statuses mathcing the title requested
            Assert.Throws(typeof(JobDoesNotExistException),
                GetNoneExistingJobStatus);


            testSubject.AddJob(validStatusJob);

            //Save statuses for later comparison
            Status validStatusJobInitialStatus = testSubject.GetJobStatus(validStatusJob.Title);

            // initial statuses before job runs should
            // be close to equal (not same datetime) as default statuses.
            Assert.IsFalse(new Status().Changed(
                validStatusJobInitialStatus));

            //Change the statuses returned by the mock job
            status.Code = Status.StatusCode.Red;

            //Run jobs so that statuses change.
            testSubject.RunJobs();
            Thread.Sleep(10);
            testSubject.WaitForAllStartedJobsToFinish();

            Assert.IsTrue(validStatusJobInitialStatus.Changed(
                testSubject.GetJobStatus(validStatusJob.Title)));

            Assert.IsTrue(status.Equals(testSubject.GetJobStatus(validStatusJob.Title)));

        }

        /// <summary>
        /// Tests the method GetJobStatus()
        /// with invalid values
        /// </summary>
        [Test]
        public void TestGetJobStatusWithInvalidStatus()
        {
            IJob nullStatusJob = new Mocks.JobMock(null);
            nullStatusJob.Title = "null";

            Assert.Throws(typeof(JobDoesNotExistException),
                GetNoneExistingJobStatus);

            testSubject.AddJob(nullStatusJob);

            //Save statuses for later comparison
            Status nullStatusJobInitialStatus = testSubject.GetJobStatus(nullStatusJob.Title);

            // initial statuses before job runs should
            // be close to equal (not same datetime) as default statuses.
            Assert.IsFalse(new Status().Changed(
               nullStatusJobInitialStatus));

            //Run jobs should change statuses to unknown with resource specific text.
            testSubject.RunJobs();
            Thread.Sleep(10);
            testSubject.WaitForAllStartedJobsToFinish();

            Status jobStatus = testSubject.GetJobStatus(nullStatusJob.Title);
            Assert.IsTrue(jobStatus.Changed(nullStatusJobInitialStatus));
        }

        /// <summary>
        /// Wrapper function. Matches TestDelegate.
        /// </summary>
        public void GetNoneExistingJobStatus()
        {
            Status status = testSubject.GetJobStatus("noneExistinTitle");
        }

        #region TestEvents

        /// <summary>
        /// Test JobFinishedEvent
        /// </summary>
        [Test]
        public void TestJobFinishedEvent()
        {
            jobFinshedEventHandlerJobTitle = String.Empty;

            testSubject.JobFinished += Scheduler_JobFinished;
            testSubject.AddJob(job);
            testSubject.RunJobs();
            Thread.Sleep(10);
            testSubject.WaitForAllStartedJobsToFinish();
            Assert.AreEqual(job.Title, jobFinshedEventHandlerJobTitle);
        }

        private string jobFinshedEventHandlerJobTitle;

        /// <summary>
        /// Event handler for job finished. Used in TestJobFinishedEvent()
        /// </summary>
        private void Scheduler_JobFinished(object sender, JobFinishedEventArgs args)
        {
            jobFinshedEventHandlerJobTitle = args.JobTitle;
        }

        /// <summary>
        /// Test StatusChanged Event
        /// </summary>
        [Test]
        public void TestStatusChangedEvent()
        {
            statusChangedEventHandlerJobTitle = String.Empty;

            Status status = new Status(Status.StatusCode.Green);
            job = new Mocks.JobMock(status);
            job.Title = "mock";

            testSubject.StatusChanged += Scheduler_StatusChanged;
            testSubject.AddJob(job);

            // Assert initial state
            Assert.AreEqual(Status.StatusCode.Unknown, testSubject.GetJobStatus(job.Title).Code);

            // run updates default state of unknown to mock job state green
            testSubject.RunJobs();

            Thread.Sleep(10);

            //wait for all jobs to finish
            testSubject.WaitForAllStartedJobsToFinish();

            // Assert new state
            Assert.AreEqual(status.Code, testSubject.GetJobStatus(job.Title).Code);

            // Assert event handler has been called
            Assert.AreEqual(job.Title, statusChangedEventHandlerJobTitle);
        }

        private string statusChangedEventHandlerJobTitle;

        /// <summary>
        /// Event handler for statuses changed. Used in TestStatusChangedEvent()
        /// </summary>
        private void Scheduler_StatusChanged(object sender, StatusChangedEventArgs args)
        {
            statusChangedEventHandlerJobTitle = args.JobTitle;
        }

        #endregion //TestEvents

        #region AutomatedJobRuns

        /// <summary>
        /// Tests having scheduler running and adding job
        /// Job should be automaticly run when added.
        /// </summary>
        [Test]
        public void TestAutomatedRun_RunningWhileAddingJob()
        {
            testSubject.Start();
            
            job = new Mocks.JobMock(new Status());
            job.Title = "Mock";
            job.RunIntervalInSeconds = 300;

            Assert.AreEqual(0, ((Mocks.JobMock)job).CallsToRun);

            testSubject.AddJob(job);

            //Sleep so the job gets a change to start.
            //Otherwise the wait for all jobs won't catch it.
            Thread.Sleep(10);
            
            testSubject.WaitForAllStartedJobsToFinish();
            Assert.AreEqual(1, ((Mocks.JobMock)job).CallsToRun);
        }

        /// <summary>
        /// Tests having scheduler turned off.
        /// Add jobs and then turn start scheduler
        /// The jobs should be run automaticly.
        /// </summary>
        [Test]
        public void TestAutomatedRun_AddingJobsAndThenStart()
        {
            job = new Mocks.JobMock(new Status());
            job.Title = "Mock";
            job.RunIntervalInSeconds = 300;
            testSubject.AddJob(job);

            Assert.AreEqual(0, ((Mocks.JobMock)job).CallsToRun);

            testSubject.Start();

            //Give the job a chance to start
            // otherwise it will not have started and
            //waitForAllJobsToFinish wont have any effect
            Thread.Sleep(20);
            
            testSubject.WaitForAllStartedJobsToFinish();
            Assert.AreEqual(1, ((Mocks.JobMock)job).CallsToRun);
        }

        #endregion //AutomatedJobRuns

        #region MultipleJobsRunning

        /// <summary>
        /// Tests that the jobs run concurrently
        /// </summary>
        //[Test]
        public void TestMultipleBusyWaitingJobsRunning()
        {
            IJob job1 = new Mocks.BusyWaitingJobMock(0.1);
            IJob job2 = new Mocks.BusyWaitingJobMock(0.1);
            double totalRunTime = 0.15;

            job1.RunIntervalInSeconds = 300;
            job1.Title = "job1";
            
            job2.RunIntervalInSeconds = 300;
            job2.Title = "job2";

            testSubject.Stop();

            testSubject.AddJob(job1);
            testSubject.AddJob(job2);

            DateTime startTime = DateTime.Now;
            
            // Run jobs concurently
            testSubject.Start();
            
            // Give jobs time to get started
            Thread.Sleep(10);
            
            testSubject.Stop();

            // Ensure that they have finished
            testSubject.WaitForAllStartedJobsToFinish();

            // Assert that actual total run time is less than if they had
            // been run sequential
            Assert.Greater(startTime.AddSeconds(totalRunTime), DateTime.Now);
            Assert.AreEqual(1, ((Mocks.JobMock)job1).CallsToRun);
            Assert.AreEqual(1, ((Mocks.JobMock)job2).CallsToRun);

            // Run jobs sequential just to prove that it takes longer
            // as expected.
            startTime = DateTime.Now;
            job1.Run();
            job2.Run();
            Assert.Less(startTime.AddSeconds(totalRunTime), DateTime.Now);

        }

        /// <summary>
        /// Test should not be run as part of unit test.
        /// Two jobs prints to console at the same time for 4 seconds.
        /// Insure that the characters are mixed.
        /// </summary>
        //[Test]
        public void TestConcurencyVisible()
        {
            IJob job1 = new Mocks.PrintingJobMock("-", 4);
            IJob job2 = new Mocks.PrintingJobMock("*", 4);
            double totalRunTime = 8;

            job1.RunIntervalInSeconds = 300;
            job1.Title = "job1";

            job2.RunIntervalInSeconds = 300;
            job2.Title = "job2";

            testSubject.Stop();

            testSubject.AddJob(job1);
            testSubject.AddJob(job2);

            DateTime startTime = DateTime.Now;

            // Run jobs concurently
            testSubject.Start();

            // Give jobs time to get started
            Thread.Sleep(10);

            // Ensure that they have finished
            testSubject.WaitForAllStartedJobsToFinish();

            testSubject.Stop();

            // Assert that actual total run time is less than if they had
            // been run sequential
            Assert.Greater(startTime.AddSeconds(totalRunTime), DateTime.Now);
            Assert.AreEqual(1, ((Mocks.JobMock)job1).CallsToRun);
            Assert.AreEqual(1, ((Mocks.JobMock)job2).CallsToRun);

        }


        #endregion

    }
}
