﻿//**********************************//
// Written by: Per Böhlin
// perbo770@student.liu.se
// Copyright: (c) 2009
//
// Implementation started: July 27, 2009
// Last modified: August 12, 2009
//**********************************//

using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Drawing;
using System.ComponentModel;
using System.Threading;
using System.IO;
using System.Diagnostics;

using DataBinding;
using DataBinding.Validators;

using BSM.core;


namespace BSM
{
    /// <summary>
    /// Main application
    /// Back end to the user interface MainFrame
    /// 
    /// Important!
    /// Close() should be called before instance is left for garbage collection
    /// 
    /// Registers plugins which makes different kind of jobs
    /// available for scheduling.
    /// 
    /// Keeps track of witch trays are to be displayed in the tray application menu.
    /// 
    /// Supports asynchronious message passing with UI-thread using
    /// a the BackgroundWorker passed in its constructor
    /// 
    /// This class is multithreaded in the sense that it creates and run
    /// jobs in separate threads. See IScheduler and Scheduler for more
    /// information.
    /// 
    /// It communicates the result of jub runs to UI-thread 
    /// via BackgroundWorker since they run in different thread compartments.
    /// 
    /// Run(object sender, DoWorkEventArgs e) - method
    /// The most important and perhaps most difficult to understand method.
    /// Run should be called By the BackgroundWorker passed in the BsmApp's constructor.
    /// This will cause the Run method to run in a separate thread, other than the
    /// UI-thread but stil allows message passing back to the UI-thread using the
    /// Background workers ReportProgress method.
    /// When interacting with the parts of the BsmApp that runs on a different thread
    /// the UI-thread should use the Ask-methods:
    ///     AskRefresh, AskStartScheduler, AskStopScheduler
    /// These methods are thread safe and can therefore be called by the UI-thread
    /// even though it is also accessed by the background thread.
    /// The Run-method is just a message loop, waiting for messages (through the ask-methods).
    /// When a message-flag is set, the loop dispatches and does some work and then returns
    /// to the loop.
    /// It only exits the loop when BackgroundWorker requests it or the Close() method is called.
    /// 
    /// General Usage:
    ///     Plugins:
    ///         PluginNames
    ///         Plugins
    /// 
    ///     Manage jobs with: CreateJob(), RemoveJob()
    /// 
    ///     Get Scheduler information with:
    ///         IsRunning, JobTitles, GetJobExist()
    /// 
    ///     Get Job Properties with:
    ///         DisplayJobInTray(), GetJobRunInterval()
    ///         GetJobTypeName(), GetJobEnabled(),
    ///     
    ///     Set Job Properties with:
    ///         AddJobToTray(), RemoveJobFromTray(),
    ///         SetJobEnabledProperty(), SetJobRunInterval,
    ///         SetNewJobTitle(),
    ///         
    ///     Save/Load Application settings:
    ///         Uses default file names from Settings
    ///         SaveApplicationSettings()
    ///         LoadApplicationSettings()
    /// 
    ///     See also IScheduler, Scheduler for better
    ///     understanding of the schedulings innerworking.
    /// 
    /// 
    /// </summary>
    class BsmApp
    {
        #region Fields
        const int maxNrOfJobsInTray = 4;

        //settings of plugins
        IDictionary<string, IPlugin> plugins;

        //Jobs displayed in tray application menu
        UniqueList<string> jobsInTray;

        //scheduled jobs
        IScheduler scheduler;

        //Application settings
        BsmApplicationSettings settings;

        //BackgroundWorker component to communicate with
        //UI thread
        BackgroundWorker threadSyncher;

        //Debug log
        TextWriter log;

        //Message flags so that the UI can ask for actions
        bool askRefresh;
        bool askStopScheduler;
        bool askStartScheduler;
        bool askCloseApp;
        object refreshLock;
        object startLock;
        object stopLock;
        object closeLock;

        #endregion //Fields

        #region Constructors

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="threadSyncher">BackgroundWorker running in UI-thread</param>
        public BsmApp(BackgroundWorker threadSyncher)
        {
            this.threadSyncher = threadSyncher;
            InitMessageFlags();
            LoadPlugins();
            InitScheduler();
            InitJobsInTray();
            InitApplicationSettings();
            InitLog();
        }

        /// <summary>
        /// Initialize fields used for message passing
        /// </summary>
        private void InitMessageFlags()
        {
            askRefresh = false;
            askStartScheduler = false;
            askStopScheduler = false;
            askCloseApp = false;
            refreshLock = new object();
            startLock = new object();
            stopLock = new object();
            closeLock = new object();
        }

        /// <summary>
        /// Initialize default application settings
        /// </summary>
        private void InitApplicationSettings()
        {
            try
            {
                settings = new BsmApplicationSettings();
                settings.DisplayMessageOnJobFinished =
                    Properties.Settings.Default.DefaultDisplayMessageOnJobFinished;
                settings.DisplayMessageOnJobStatusChanged =
                    Properties.Settings.Default.DefaultDisplayMessageOnJobStatusChanged;
                settings.DisplayMessageOnError =
                    Properties.Settings.Default.DefaultDisplayMessageOnError;
                settings.MainFormActionOnClick = (ActionOnClick)Properties.Settings.Default.DefaultMainFormActionOnClick;
                settings.TrayActionOnClick = (ActionOnClick)Properties.Settings.Default.DefaultTrayActionOnClick;
            }
            catch
            {
                DisplayMessage(new BsmMessage("", Properties.Resources.ErrorCaption,
                    Properties.Resources.InitSettingsErrorMessage, BsmMessageTypes.Exception));
            }
        }

        /// <summary>
        /// Initialize fields associated with keeping track of jobs displayed in the tray application menu
        /// </summary>
        private void InitJobsInTray()
        {
            jobsInTray = new UniqueList<string>(maxNrOfJobsInTray);
        }

        /// <summary>
        /// Initialize scheduler
        /// </summary>
        private void InitScheduler()
        {
            scheduler = new Scheduler();
            scheduler.JobFinished += this.Scheduler_JobFinished;
            scheduler.StatusChanged += this.Scheduler_StatusChanged;
            scheduler.Start();
        }

        /// <summary>
        /// Init Log file, if LOG flag is set.
        /// </summary>
        [Conditional("LOG")]
        private void InitLog()
        {
            log = new StreamWriter(Properties.Settings.Default.BsmApplicationDebugLogFilePath, false);
        }

        /// <summary>
        /// Load available plugins
        /// </summary>
        private void LoadPlugins()
        {
            plugins = PluginLoader.Load(Path.GetFullPath(Properties.Settings.Default.PluginDirectory));
        }

        #endregion //Constructors

        #region Properties

        /// <summary>
        /// String-array with the names of all registered plugins
        /// </summary>
        public string[] PluginNames
        {
            get
            {
                string[] pluginNames = new string[plugins.Count];
                plugins.Keys.CopyTo(pluginNames, 0);
                return pluginNames;
            }
        }

        /// <summary>
        /// Array with all registered plugins
        /// </summary>
        public IPlugin[] Plugins
        {
            get
            {
                IPlugin[] result = new IPlugin[plugins.Count];
                plugins.Values.CopyTo(result, 0);
                return result;
            }
        }

        /// <summary>
        /// String-array with all scheduled job titles
        /// </summary>
        public string[] JobTitles
        {
            get { return scheduler.JobTitles; }
        }

        /// <summary>
        /// True if scheduler is running
        /// Read only property
        /// </summary>
        public bool IsRunning
        {
            get { return scheduler.IsRunning; }
        }

        /// <summary>
        /// Action to take when a job item is clicked on the main form.
        /// </summary>
        public ActionOnClick MainFormActionOnClick
        {
            get { return settings.MainFormActionOnClick; }
        }

        /// <summary>
        /// Action to take when a job item is clicked on the tray context menu
        /// </summary>
        public ActionOnClick TrayActionOnClick
        {
            get { return settings.TrayActionOnClick; }
        }

        #endregion //Properties

        #region JobMethods

        /// <summary>
        /// Create new job and add to scheduler
        /// Require user interaction
        /// </summary>
        /// /// <param name="owner">owner of form</param>
        /// <returns>true if successful. false if user canceled</returns>
        public bool CreateJob(IWin32Window owner)
        {
            JobFrame form = new JobFrame(this.PluginNames, this.JobTitles,
                TraySpotsAvailable());
            if (form.ShowDialog() == DialogResult.OK)
            {
                string title = form.Title;
                IJob job = CreateJob(form.JobType, title, form.RunInterval, true);
                if (ConfigureJob(job, owner))
                {
                    AddJob(job, form.DisplayInTray);
                }

                return true;
            }
            return false;
        }

        /// <summary>
        /// Add job to scheduler and marked as display in tray if
        /// displayInTray is true
        /// </summary>
        /// <param name="job">job</param>
        /// <param name="displayInTray">bool, if true, job is marked to be displayed in tray</param>
        private void AddJob(IJob job, bool displayInTray)
        {
            scheduler.AddJob(job);
            if (displayInTray)
                AddJobToTray(job.Title);
        }

        /// <summary>
        /// Create a new job instance using the corresponding plugin
        /// throws:
        ///     ArgumentException
        /// </summary>
        /// <param name="type">job type/plugin name</param>
        /// <param name="title">job title</param>
        /// <param name="runInterval">run interval, in seconds</param>
        /// <param name="enabled">bool : value of the Enabled property</param>
        /// <returns>the created job</returns>
        private IJob CreateJob(string type, string title, int runInterval, bool enabled)
        {
            IPlugin plugin = plugins[type];
            if (plugin == null)
                throw new ArgumentException("Invalid plugin type");

            IJob job = plugin.CreateInstance();
            SetNewJobTitle(job, title);
            SetJobRunInterval(job, runInterval);
            SetJobEnabledProperty(job, enabled);
            return job;
        }

        /// <summary>
        /// Remove job
        /// The job's settings file will be deleted.
        /// </summary>
        /// <param name="title">job title</param>
        public void RemoveJob(string title)
        {
            try
            {
                DeleteJobSettings(title);
                RemoveJobWitoutDeletingSettings(title);
            }
            catch (BsmException ex)
            {
                WriteExceptionToLog(ex);
                DisplayMessage(new BsmMessage(title, title, ex.Message, BsmMessageTypes.Exception));
            }
        }

        private void RemoveJobWitoutDeletingSettings(string title)
        {
            scheduler.RemoveJob(title);
            jobsInTray.Remove(title);
        }

        /// <summary>
        /// Get job by title
        /// throws BsmException
        /// </summary>
        /// <param name="title">job title</param>
        /// <returns>job</returns>
        private IJob GetJob(string title)
        {
            try
            {
                return scheduler.GetJob(title);
            }
            catch (KeyNotFoundException ex)
            {
                throw new BsmException(Properties.Resources.JobNotFoundErrorMessage, ex);
            }
        }

        public bool GetJobExists(string jobTitle)
        {
            return scheduler.GetJobExists(jobTitle);
        }

        /// <summary>
        /// Get statuses for a specific job
        /// Returns the statuses of the last run
        /// </summary>
        /// <param name="title">job title</param>
        /// <returns>statuses of last run</returns>
        public Status GetJobStatus(string title)
        {
            return scheduler.GetJobStatus(title);
        }

        /// <summary>
        /// Get the name of the job type for a given job by title
        /// </summary>
        /// <param name="title">job title</param>
        /// <returns>Job type name</returns>
        public string GetJobTypeName(string title)
        {
            try
            {
                return GetJob(title).TypeName;
            }
            catch (BsmException ex)
            {
                WriteExceptionToLog(ex);
                DisplayMessage(new BsmMessage(title, title, ex.Message, BsmMessageTypes.Exception));
                return String.Empty;
            }
        }

        /// <summary>
        /// Gets the job enabled property statuses, by title
        /// </summary>
        /// <param name="title">job title</param>
        /// <returns>true if the job is enabled</returns>
        public bool GetJobEnabled(string title)
        {
            try
            {
                return GetJob(title).Enabled;
            }
            catch (BsmException ex)
            {
                WriteExceptionToLog(ex);
                DisplayMessage(new BsmMessage(title, title, ex.Message, BsmMessageTypes.Exception));
                return false;
            }
        }

        /// <summary>
        /// Get the run interval for a given job, by title
        /// </summary>
        /// <param name="title">job title</param>
        /// <returns>run interval in seconds</returns>
        public int GetJobRunInterval(string title)
        {
            try
            {
                return GetJob(title).RunIntervalInSeconds;
            }
            catch (BsmException ex)
            {
                WriteExceptionToLog(ex);
                DisplayMessage(new BsmMessage(title, title, ex.Message, BsmMessageTypes.Exception));
                return Properties.Settings.Default.DefaultRunInterval;
            }
        }

        /// <summary>
        /// Set new title for a job.
        /// Will cause the job to reregister with the scheduler
        /// </summary>
        /// <param name="currentTitle">current job title</param>
        /// <param name="newTitle">new job title</param>
        /// <returns>true if successful</returns>
        public bool SetNewJobTitle(string currentTitle, string newTitle)
        {
            try
            {
                DataField<string> jobTitle = new DataField<string>(
                    new JobTitleValidator(JobTitles),
                    null, newTitle);

                if (!jobTitle.IsValid())
                {
                    DisplayMessage(new BsmMessage(currentTitle,
                        String.Format(Properties.Resources.SetNewJobTitleFailedCaptionFormat,
                            currentTitle),
                        Properties.Resources.InvalidInputJobTitle, BsmMessageTypes.General));
                    return false;
                }

                IJob job = GetJob(currentTitle);
                bool displayInTray = DisplayJobInTray(currentTitle);
                //Always remove job before modifying it.
                RemoveJob(currentTitle);
                SetNewJobTitle(job, newTitle);
                SaveJobSettings(job);
                AddJob(job, displayInTray);
                return true;
            }
            catch (BsmException ex)
            {
                WriteExceptionToLog(ex);
                DisplayMessage(new BsmMessage(currentTitle, currentTitle, ex.Message, BsmMessageTypes.Exception));
                return false;
            }
        }

        /// <summary>
        /// set new job title. Job can not be scheduled
        /// </summary>
        /// <param name="job">job. can not be scheduled</param>
        /// <param name="newTitle">new job title</param>
        private void SetNewJobTitle(IJob job, string newTitle)
        {
            if (scheduler.GetJobExists(job.Title))
                throw new ModifyScheduledJobException();
            job.Title = newTitle;
        }

        /// <summary>
        /// Set the run interval for a given job, by title
        /// Run interval is given in seconds
        /// </summary>
        /// <param name="title">job title</param>
        /// <param name="runInterval">run interval in seconds</param>
        public void SetJobRunInterval(string title, int runInterval)
        {
            try
            {
                IJob job = GetJob(title);
                scheduler.RemoveJob(title);
                SetJobRunInterval(job, runInterval);
                scheduler.AddJob(job);
            }
            catch (BsmException ex)
            {
                WriteExceptionToLog(ex);
                DisplayMessage(new BsmMessage(title, title, ex.Message, BsmMessageTypes.Exception));
            }
        }

        private void SetJobRunInterval(IJob job, int runInterval)
        {
            if (scheduler.GetJobExists(job.Title))
                throw new ModifyScheduledJobException();

            job.RunIntervalInSeconds = runInterval;
        }

        /// <summary>
        /// Set the Enabled property for a job, given by title
        /// </summary>
        /// <param name="title">job title</param>
        public void SetJobEnabledProperty(string title, bool enabled)
        {
            try
            {
                IJob job = GetJob(title);
                if (job.Enabled != enabled)
                {
                    scheduler.RemoveJob(title);
                    SetJobEnabledProperty(job, enabled);
                    scheduler.AddJob(job);
                }
            }
            catch (BsmException ex)
            {
                WriteExceptionToLog(ex);
                DisplayMessage(new BsmMessage(title, title, ex.Message, BsmMessageTypes.Exception));
            }
        }

        /// <summary>
        /// Set job enabled property. Job should not be scheduled
        /// </summary>
        /// <param name="job">job, should not be scheduled</param>
        /// <param name="enabled">bool : new value of enabled property</param>
        private void SetJobEnabledProperty(IJob job, bool enabled)
        {
            if (scheduler.GetJobExists(job.Title))
                throw new ModifyScheduledJobException();

            job.Enabled = enabled;
        }

        /// <summary>
        /// Configure a job, by title
        /// </summary>
        /// <param name="title">job title</param>
        public void ConfigureJob(string title, IWin32Window owner)
        {
            try
            {
                IJob job = GetJob(title);
                bool displayInTray = DisplayJobInTray(title);
                RemoveJobWitoutDeletingSettings(title);
                ConfigureJob(job, owner);
                AddJob(job, displayInTray);
            }
            catch (BsmException ex)
            {
                WriteExceptionToLog(ex);
                DisplayMessage(new BsmMessage(title, title, ex.Message, BsmMessageTypes.Exception));
            }
        }

        /// <summary>
        /// Configure a job. The job should not be scheduled
        /// </summary>
        /// <param name="job">job being configured. The job should not be scheduled.</param>
        /// <param name="owner">owner/parent to the configuration window</param>
        /// <returns>true if job was configured</returns>
        private bool ConfigureJob(IJob job, IWin32Window owner)
        {
            if (scheduler.GetJobExists(job.Title))
                throw new ModifyScheduledJobException();

            if (job.Configure(owner))
            {
                job.SaveSettings(GetSettingsFilePath(job));
                return true;
            }
            return false;
        }

        /// <summary>
        /// Launch the application associated with the job
        /// </summary>
        /// <param name="title">job title</param>
        public void LaunchJobApplication(string title)
        {
            try
            {
                IJob job = GetJob(title);
                Thread thread = new Thread(new ThreadStart(job.Launch));
                thread.TrySetApartmentState(ApartmentState.MTA);

                //Otherwise the BSA application won't really quit if thread still alive
                //when application exits.
                //Otherwise could a ill implemented job.Launch force the application process
                //to keep running even though Window and tray is gone.
                thread.IsBackground = true; 

                thread.Name = "Launch " + title + " Application";
                thread.Start();
                Thread.Sleep(0);
            }
            catch (BsmException ex)
            {
                WriteExceptionToLog(ex);
                // Catches JobDoesNotExcist from GetJob()
                // Can't really do anything about exceptions in the new thread since it runs on
                // a separate stack compartment.
                DisplayMessage(new BsmMessage(title, title, ex.Message, BsmMessageTypes.Exception));
            }
        }

        #endregion //Job methods

        #region TrayMethods

        /// <summary>
        /// Add job to be displayed in tray application menu
        /// </summary>
        /// <param name="title">job title</param>
        /// <returns>true if successful</returns>
        public bool AddJobToTray(string title)
        {
            return jobsInTray.Add(title);
        }

        /// <summary>
        /// Returns true if job is displayed in tray application menu
        /// </summary>
        /// <param name="title">job title</param>
        /// <returns>true if job is displayed in tray application menu</returns>
        public bool DisplayJobInTray(string title)
        {
            return jobsInTray.Contains(title);
        }

        /// <summary>
        /// Returns true if there are stil spots available for jobs
        /// being displayed in the tray application menu
        /// </summary>
        /// <returns>true if available</returns>
        public bool TraySpotsAvailable()
        {
            return (jobsInTray.Count < maxNrOfJobsInTray);
        }

        /// <summary>
        /// Sorted string-array of  job titles of the jobs displayed in tray application menu
        /// Read-only property
        /// </summary>
        public string[] JobsInTray
        {
            get
            {
                jobsInTray.Sort();
                string[] result = new string[jobsInTray.Count];
                jobsInTray.CopyTo(result);
                
                return result;
            }
        }

        /// <summary>
        /// Remove job from being displayed in tray application menu
        /// </summary>
        /// <param name="title">job title</param>
        public void RemoveJobFromTray(string title)
        {
            jobsInTray.Remove(title);
        }

        #endregion //TrayMethods

        #region Run

        /// <summary>
        /// Special method. Called by BackgroundWorker to start the
        /// application. Without it, the BsmApp cannot communicate
        /// with a User interface thread due to the threads compartment model
        /// e.Result = false when exits because of error. Otherwise true
        /// </summary>
        public void Run(object sender, DoWorkEventArgs e)
        {
            try
            {
                while (!threadSyncher.CancellationPending && !askCloseApp)
                {
                    //Check for message flags;

                    CheckMessageFlagsAndDoAction();

                    //Let the thread yield
                    //Scheduler's timers will fire off and call
                    //threadSynchers progressUpdate to notify
                    //when event fires.
                    System.Threading.Thread.Sleep(20);
                }
                e.Cancel = true;
                e.Result = true;
            }
            catch (Exception ex)
            {
                DisplayMessage(new BsmException(String.Format(Properties.Resources.RunExceptionFormat, ex.Message), ex));
                e.Cancel = true;
                e.Result = false;
            }
        }

        /// <summary>
        /// Checks if the user interface thread has asked
        /// for specific actions using the Message flags:
        /// Refresh, StartScheduler, StopScheduler
        /// If a request has been made, the corresponding
        /// action is done.
        /// When the action has been performed, a callback
        /// is issued to the BackgroundWorker, sending
        /// a BsmMessage.
        /// Thread safe
        /// </summary>
        private void CheckMessageFlagsAndDoAction()
        {
            CheckAndDoRefresh();
            ScheckAndDoStartScheduler();
            CheckAndDoStopScheduler();
        }
        
        /// <summary>
        /// Message flag: Refresh
        /// Run all jobs in scheduler
        /// Results in callback to Backgroundworker
        /// with a BsmMessage
        /// Thread safe
        /// </summary>
        private void CheckAndDoRefresh()
        {
            lock (refreshLock)
            {
                if (askRefresh)
                {
                    scheduler.RunJobs();
                    askRefresh = false;
                    DisplayMessage(new BsmMessage(BsmMessageTypes.Refresh));
                }
            }
        }

        /// <summary>
        /// Message flag: StartScheduler
        /// Startas the scheduler
        /// Results in callback to Backgroundworker
        /// with a BsmMessage
        /// Thread safe
        /// </summary>
        private void ScheckAndDoStartScheduler()
        {
            lock (startLock)
            {
                if (askStartScheduler)
                {
                    scheduler.Start();
                    askStartScheduler = false;
                    DisplayMessage(new BsmMessage(BsmMessageTypes.StartingScheduler));
                }
            }
        }

        /// <summary>
        /// Message flag: Refresh
        /// Stopds the scheduler
        /// Results in callback to Backgroundworker
        /// with a BsmMessage
        /// Thread safe
        /// </summary>
        private void CheckAndDoStopScheduler()
        {
            lock (stopLock)
            {
                if (askStopScheduler)
                {
                    scheduler.Stop();
                    askStopScheduler = false;
                    DisplayMessage(new BsmMessage(BsmMessageTypes.StoppingScheduler));
                }
            }
        }

        /// <summary>
        /// Message flag: Refresh
        /// Requests a refresh of all scheduled jobs
        /// Called by UI-thread
        /// Thread safe
        /// </summary>
        public void AskRefresh()
        {
            lock (refreshLock)
            {
                askRefresh = true;
            }
        }

        /// <summary>
        /// Message flag: Start Scheduler
        /// Requests that the scheduler is started
        /// Called by UI-thread
        /// Thread safe
        /// </summary>
        public void AskStartScheduler()
        {
            lock (startLock)
            {
                askStartScheduler = true;
            }
        }

        /// <summary>
        /// Message flag: Start Scheduler
        /// Requests that the scheduler is stopped
        /// Called by UI-thread
        /// Thread safe
        /// </summary>
        public void AskStopScheduler()
        {
            lock (stopLock)
            {
                askStopScheduler = true;
            }
        }

        #endregion //Run

        /// <summary>
        /// Should be called before released for garbage collection.
        /// </summary>
        public void Close()
        {
            lock (closeLock)
            {
                askCloseApp = true;
            }
            scheduler.JobFinished -= Scheduler_JobFinished;
            scheduler.StatusChanged -= Scheduler_StatusChanged;
            scheduler.Close();
            scheduler.WaitForAllStartedJobsToFinish();
            CloseLog();
            GC.Collect();
            
        }

        /// <summary>
        /// Close the log if LOG flag is set
        /// </summary>
        [Conditional("LOG")]
        private void CloseLog()
        {
            lock (log)
            {
                log.Close();
            }
        }

        #region ConfigurationAndSettings

        /// <summary>
        /// Configure the Bsm Application Settings
        /// Require user interaction
        /// </summary>
        public void ConfigureApplication(IWin32Window owner)
        {
            try
            {
                settings.Configure(owner);
            }
            catch (BsmException ex)
            {
                WriteExceptionToLog(ex);
                DisplayMessage(ex);
            }
        }

        /// <summary>
        /// Load application settings from Default file location
        /// Done so in its own thread.
        /// </summary>
        public void LoadApplicationSettings()
        {
            try
            {
                string filePath = GetApplicationSettingsFilePath();
                settings = BsmApplicationSettings.LoadFromFile(filePath);
                
                if (settings.Running)
                    scheduler.Start();
                else
                    scheduler.Stop();

                foreach (JobListing jobListing in settings.JobList)
                {
                    IJob job = CreateJob(jobListing.TypeName, jobListing.Title, jobListing.RunInterval, jobListing.Enabled);

                    LoadJobSettings(job);
                    AddJob(job, jobListing.DisplayInTray);
                }
            }
            catch (BsmException ex)
            {
                WriteExceptionToLog(ex);
                DisplayMessage(ex);
            }
        }

        /// <summary>
        /// Returns the full file path to the Application settings file.
        /// </summary>
        /// <returns></returns>
        private static string GetApplicationSettingsFilePath()
        {
            return Path.GetFullPath(String.Format("{0}/{1}",
                                Properties.Settings.Default.SettingsDirectory,
                                Properties.Settings.Default.BsmApplicationSettingsFileName));
        }

        /// <summary>
        /// Save application settings to default file location
        /// </summary>
        public void SaveApplicationSettings()
        {
            try
            {
                string filePath = GetApplicationSettingsFilePath();
                settings.ClearJobListing();
                foreach (IJob job in scheduler.Jobs)
                {
                    settings.AddJobListing(new JobListing(job, jobsInTray.Contains(job.Title)));
                }
                settings.Running = scheduler.IsRunning;

                BsmApplicationSettings.SaveToFile(filePath, settings);
            }
            catch (BsmException ex)
            {
                WriteExceptionToLog(ex);
                DisplayMessage(ex);
            }
        }

        #endregion //ConfigurationAndSettings

        /// <summary>
        /// If BackgroundWorking is set,
        /// messages will be forwarded to that.
        /// It is through this mechanism BsmApp
        /// communicates with UI-threads.
        /// </summary>
        /// <param name="message">message to be displayed</param>
        public void DisplayMessage(BsmMessage message)
        {

            switch (message.MessageType)
            {
                case BsmMessageTypes.Exception:
                    message.Display = settings.DisplayMessageOnError;
                    break;
                case BsmMessageTypes.JobStatusChanged:
                    message.Display = settings.DisplayMessageOnJobStatusChanged;
                    break;
                case BsmMessageTypes.JobFinished:
                    message.Display = settings.DisplayMessageOnJobFinished;
                    break;
                case BsmMessageTypes.General:
                    message.Display = true;
                    break;
                default:
                    message.Display = false;
                    break;
            }
            if ((threadSyncher != null) && threadSyncher.IsBusy && (!threadSyncher.CancellationPending))
            {
                threadSyncher.ReportProgress(0, message);
            }
        }

        /// <summary>
        /// Wrapper method for displaying a BsmException message
        /// </summary>
        /// <param name="exception">BsmException</param>
        private void DisplayMessage(BsmException exception)
        {
            DisplayMessage(new BsmMessage("", Properties.Resources.ErrorCaption, exception.Message, BsmMessageTypes.Exception));
        }

        /// <summary>
        /// Wrapper method for displaying a BsmMessage based on a UIException
        /// </summary>
        /// <param name="exception">UIException</param>
        public void DisplayMessage(UIException exception)
        {
            DisplayMessage(new BsmMessage("", Properties.Resources.ErrorCaption, exception.Message, BsmMessageTypes.UIException));
        }

        #region EventHandlers

        /// <summary>
        /// Event handler for scheduler's JobFinished event
        /// </summary>
        private void Scheduler_JobFinished(object sender, JobFinishedEventArgs args)
        {
            try
            {
                Status status = GetJobStatus(args.JobTitle);
                BsmMessage message = new BsmMessage(args.JobTitle,
                    args.JobTitle, status.Text, BsmMessageTypes.JobFinished);

                DisplayMessage(message);
            }
            catch (BsmException ex)
            {
                WriteExceptionToLog(ex);
                DisplayMessage(ex);
            }
        }

        /// <summary>
        /// Event handler for scheduler's StatusChanged event
        /// </summary>
        private void Scheduler_StatusChanged(object sender, StatusChangedEventArgs args)
        {
            Status status = GetJobStatus(args.JobTitle);
            BsmMessage message = new BsmMessage(args.JobTitle,
                args.JobTitle, status.Text, BsmMessageTypes.JobStatusChanged);

            DisplayMessage(message);
        }

        #endregion //EventHandlers

        #region JobSettings

        /// <summary>
        /// Save settings for a specific job
        /// Not thread safe. If job runs changes settings
        /// this could save inconsistant values
        /// </summary>
        /// <param name="job">job</param>
        private void SaveJobSettings(IJob job)
        {
            try
            {
                // can not lock on job since
                // job run can have long execution time and
                // this would lock the UI-thread
                 job.SaveSettings(GetSettingsFilePath(job));
            }
            catch (BsmException ex)
            {
                WriteExceptionToLog(ex);
                DisplayMessage(new BsmMessage(job.Title, job.Title, ex.Message, BsmMessageTypes.Exception));
            }
        }

        /// <summary>
        /// Load settings for a specific job
        /// Job should not be scheduled
        /// </summary>
        /// <param name="job">job, should not be scheduled</param>
        private void LoadJobSettings(IJob job)
        {
            try
            {
                if (scheduler.GetJobExists(job.Title))
                    throw new ModifyScheduledJobException();

                job.LoadSettings(GetSettingsFilePath(job));
            }
            catch (BsmException ex)
            {
                WriteExceptionToLog(ex);
                DisplayMessage(new BsmMessage(job.Title, job.Title, ex.Message, BsmMessageTypes.Exception));
            }
        }

        /// <summary>
        /// Delete settings for a specific job given by title
        /// </summary>
        /// <param name="title"></param>
        private void DeleteJobSettings(string title)
        {
            try
            {
                IJob job = GetJob(title);
                job.DeleteSettings(GetSettingsFilePath(job));
            }
            catch (BsmException ex)
            {
                WriteExceptionToLog(ex);
                DisplayMessage(new BsmMessage(title, title, ex.Message, BsmMessageTypes.Exception));
            }
        }

        /// <summary>
        /// Get the Settings File path for a job.
        /// </summary>
        /// <param name="job">job</param>
        /// <returns>file path</returns>
        private string GetSettingsFilePath(IJob job)
        {
            return Path.GetFullPath(String.Format("{0}/{1}.{2}.xml", 
                Properties.Settings.Default.SettingsDirectory, job.TypeName, job.Title));
        }

        #endregion //JobSettings

        #region Debug

        /// <summary>
        /// Write a debug report to file
        /// Contains the internal state of the application.
        /// </summary>
        public void WriteDebugReport()
        {
            
            try
            {
                string filePath = Properties.Settings.Default.BsmApplicationDebugReportFilePath;
                using (TextWriter writer = new StreamWriter(filePath))
                {
                    WriteDebugReport(writer);           
                    writer.Close();
                    DisplayMessage(new BsmMessage("", Properties.Resources.DebugReportCaption,
                        Properties.Resources.DebugReportMessage + " " + filePath, BsmMessageTypes.General));
                }
            }
            catch (BsmException ex)
            {
                WriteExceptionToLog(ex);
                DisplayMessage(new BsmMessage("", Properties.Resources.ErrorCaption, ex.Message, BsmMessageTypes.Exception));
            }
        }

        /// <summary>
        /// Writes debug report to provided writer
        /// </summary>
        /// <param name="writer">Textwriter written to</param>
        private void WriteDebugReport(TextWriter writer)
        {
            DebugReporter reporter = new DebugReporter(writer);
            reporter.WriteHeader("Build System Monitor Debug Report");
            reporter.WriteAssemblyInfo();
            reporter.WriteArray(PluginNames, "Plugins");
            reporter.WriteArray(JobsInTray, "Jobs In Tray");
            reporter.WriteArray(JobTitles, "Scheduled Jobs");
            reporter.WriteHeader("Fields");
            reporter.WriteValuePair("askRefresh", askRefresh.ToString());
            reporter.WriteValuePair("askStartScheduler", askStartScheduler.ToString());
            reporter.WriteValuePair("askStopScheduler", askStopScheduler.ToString());
            reporter.WriteParagraph("");
            reporter.WriteResources(typeof(Properties.Resources));
        }

        /// <summary>
        /// Log an BsmException to the application log if the LOG flag is set at build time
        /// Thread safe
        /// </summary>
        /// <param name="exception"></param>
        [Conditional("LOG")]
        public void WriteExceptionToLog(BsmException exception)
        {
            lock (log)
            {
                log.WriteLine(DateTime.Now.ToString() + " " + exception.ToString());
                log.WriteLine("  " + exception.StackTrace);
                log.Flush();
            }
        }

        #endregion //Debug
    }
}
