﻿//**********************************//
// Written by: Per Böhlin
// perbo770@student.liu.se
// Copyright: (c) 2009
//
// Implementation started: July 27, 2009
// Last modified: August 12, 2009
//**********************************//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using BSM.core;
using DataBinding;

namespace BSM
{
    public partial class MainFrame : Form
    {
        // Almost all logic is in this class.
        // Parts of it is accessed asynchronious using
        // a BackgroundWorker.
        BsmApp bsmApp;

        // used in Drag n Drop
        bool mouseDown;

        // used in label edit
        string labelEdited;

        // list of job items on the tray application context menu
        List<ToolStripMenuItem> jobsOnTrayMenu;

        // Icon list
        Dictionary<string, Icon> trayIconList;

        #region Constructors

        /// <summary>
        /// Constructor
        /// </summary>
        public MainFrame()
        {
            // GUI Designer initilizations
            InitializeComponent();

            // Field initilaizations
            FieldInitilaizations();

            //Custom initilizations
            InitializeGUI(bsmApp.Plugins);

            //Start background service
            threadSyncher.RunWorkerAsync();

            //Load Application settings
            bsmApp.LoadApplicationSettings();

            //Update form
            UpdateForm();
        }

        /// <summary>
        /// Field initilaizations
        /// </summary>
        private void FieldInitilaizations()
        {
            mouseDown = false;
            labelEdited = String.Empty;
            bsmApp = new BsmApp(threadSyncher);
            jobsOnTrayMenu = new List<ToolStripMenuItem>();
            trayIconList = new Dictionary<string, Icon>();
        }

        #endregion Constructors

        #region InitAndUpdate

        /// <summary>
        /// Custom initilizations of user interface
        /// </summary>
        /// <param name="plugins">array of plugins available to the application</param>
        private void InitializeGUI(IPlugin[] plugins)
        {
            SetEventHandlers();
            LoadImageLists(plugins);
            SetTileSize();
            InitTray();
        }

        /// <summary>
        /// Initialize tray application
        /// </summary>
        private void InitTray()
        {
            tray.Text = Application.ProductName;
        }

        /// <summary>
        /// Set forms eventhandlers
        /// </summary>
        private void SetEventHandlers()
        {
            // Closing form event handler
            this.FormClosing += MainFrame_FormClosing;
            this.KeyDown += MainFrame_StartDebugMode;

            // Event handlers for view of jobs
            this.lvwJobs.MouseDoubleClick += this.lvwJobs_MouseDoubleClick;
            this.lvwJobs.KeyDown += this.lvwJobs_KeyDown;
            this.lvwJobs.Resize += this.lvwJobs_Resize;
            this.lvwJobs.DragEnter += this.lvwJobs_DragEnter;
            this.lvwJobs.DragDrop += this.lvwJobs_DragDrop;
            this.lvwJobs.MouseDown += this.lvwJobs_MouseDown;
            this.lvwJobs.MouseMove += this.lvwJobs_MouseMove;
            this.lvwJobs.BeforeLabelEdit += this.lvwJobs_BeforeLabelEdit;
            this.lvwJobs.AfterLabelEdit += this.lvwJobs_AfterLabelEdit;

            // tray application
            tray.DoubleClick += this.tray_DoubleClick;

            // threadSyncher
            threadSyncher.DoWork += new DoWorkEventHandler(bsmApp.Run);
            threadSyncher.ProgressChanged += new ProgressChangedEventHandler(threadSyncher_EventRaized);
            //RunWorkerCompleted should not be used, since it is only invoked after
            //the form terminates
            //threadSyncher.RunWorkerCompleted
        }

        /// <summary>
        /// Load image lists
        /// State icon list: unknown, green, yellow, red
        /// Job type icon list: one for each plugin
        /// </summary>
        /// <param name="plugins">array of plugins available</param>
        private void LoadImageLists(IPlugin[] plugins)
        {
            // Load plugin icons into image settings
            LoadPluginIcons(plugins);

            // Added in the order so that index match Status.StatusCode int values
            LoadStateIcons();

            // Tray icons
            LoadTrayIcons();
        }

        /// <summary>
        /// Load the plugin specific icons to internal image list.
        /// </summary>
        /// <param name="plugins"></param>
        private void LoadPluginIcons(IPlugin[] plugins)
        {
            foreach (IPlugin plugin in plugins)
            {
                ilsJobIcons.Images.Add(plugin.Name, plugin.Icon);
            }
        }

        /// <summary>
        /// Load icons representing state to internal image list.
        /// </summary>
        private void LoadStateIcons()
        {
            ilsStateIcons.Images.Add(Status.StatusCode.Unknown.ToString(), Properties.Resources.state_unknown);
            ilsStateIcons.Images.Add(Status.StatusCode.Green.ToString(), Properties.Resources.state_green);
            ilsStateIcons.Images.Add(Status.StatusCode.Yellow.ToString(), Properties.Resources.state_yellow);
            ilsStateIcons.Images.Add(Status.StatusCode.Red.ToString(), Properties.Resources.state_red);
            
        }

        /// <summary>
        /// Load tray icons to internal list.
        /// </summary>
        private void LoadTrayIcons()
        {
            trayIconList.Add("App", Properties.Resources.BSM);
            trayIconList.Add("G", Properties.Resources.G);
            trayIconList.Add("Y", Properties.Resources.Y);
            trayIconList.Add("R", Properties.Resources.R);
            trayIconList.Add("U", Properties.Resources.U);
            trayIconList.Add("UUUU", Properties.Resources.UUUU);
            trayIconList.Add("UUUG", Properties.Resources.UUUG);
            trayIconList.Add("UUUY", Properties.Resources.UUUY);
            trayIconList.Add("UUUR", Properties.Resources.UUUR);
            trayIconList.Add("UUGU", Properties.Resources.UUGU);
            trayIconList.Add("UUGG", Properties.Resources.UUGG);
            trayIconList.Add("UUGY", Properties.Resources.UUGY);
            trayIconList.Add("UUGR", Properties.Resources.UUGR);
            trayIconList.Add("UUYU", Properties.Resources.UUYU);
            trayIconList.Add("UUYG", Properties.Resources.UUYG);
            trayIconList.Add("UUYY", Properties.Resources.UUYY);
            trayIconList.Add("UUYR", Properties.Resources.UUYR);
            trayIconList.Add("UURU", Properties.Resources.UURU);
            trayIconList.Add("UURG", Properties.Resources.UURG);
            trayIconList.Add("UURY", Properties.Resources.UURY);
            trayIconList.Add("UURR", Properties.Resources.UURR);
            trayIconList.Add("UGUU", Properties.Resources.UGUU);
            trayIconList.Add("UGUG", Properties.Resources.UGUG);
            trayIconList.Add("UGUY", Properties.Resources.UGUY);
            trayIconList.Add("UGUR", Properties.Resources.UGUR);
            trayIconList.Add("UGGU", Properties.Resources.UGGU);
            trayIconList.Add("UGGG", Properties.Resources.UGGG);
            trayIconList.Add("UGGY", Properties.Resources.UGGY);
            trayIconList.Add("UGGR", Properties.Resources.UGGR);
            trayIconList.Add("UGYU", Properties.Resources.UGYU);
            trayIconList.Add("UGYG", Properties.Resources.UGYG);
            trayIconList.Add("UGYY", Properties.Resources.UGYY);
            trayIconList.Add("UGYR", Properties.Resources.UGYR);
            trayIconList.Add("UGRU", Properties.Resources.UGRU);
            trayIconList.Add("UGRG", Properties.Resources.UGRG);
            trayIconList.Add("UGRY", Properties.Resources.UGRY);
            trayIconList.Add("UGRR", Properties.Resources.UGRR);
            trayIconList.Add("UYUU", Properties.Resources.UYUU);
            trayIconList.Add("UYUG", Properties.Resources.UYUG);
            trayIconList.Add("UYUY", Properties.Resources.UYUY);
            trayIconList.Add("UYUR", Properties.Resources.UYUR);
            trayIconList.Add("UYGU", Properties.Resources.UYGU);
            trayIconList.Add("UYGG", Properties.Resources.UYGG);
            trayIconList.Add("UYGY", Properties.Resources.UYGY);
            trayIconList.Add("UYGR", Properties.Resources.UYGR);
            trayIconList.Add("UYYU", Properties.Resources.UYYU);
            trayIconList.Add("UYYG", Properties.Resources.UYYG);
            trayIconList.Add("UYYY", Properties.Resources.UYYY);
            trayIconList.Add("UYYR", Properties.Resources.UYYR);
            trayIconList.Add("UYRU", Properties.Resources.UYRU);
            trayIconList.Add("UYRG", Properties.Resources.UYRG);
            trayIconList.Add("UYRY", Properties.Resources.UYRY);
            trayIconList.Add("UYRR", Properties.Resources.UYRR);
            trayIconList.Add("URUU", Properties.Resources.URUU);
            trayIconList.Add("URUG", Properties.Resources.URUG);
            trayIconList.Add("URUY", Properties.Resources.URUY);
            trayIconList.Add("URUR", Properties.Resources.URUR);
            trayIconList.Add("URGU", Properties.Resources.URGU);
            trayIconList.Add("URGG", Properties.Resources.URGG);
            trayIconList.Add("URGY", Properties.Resources.URGY);
            trayIconList.Add("URGR", Properties.Resources.URGR);
            trayIconList.Add("URYU", Properties.Resources.URYU);
            trayIconList.Add("URYG", Properties.Resources.URYG);
            trayIconList.Add("URYY", Properties.Resources.URYY);
            trayIconList.Add("URYR", Properties.Resources.URYR);
            trayIconList.Add("URRU", Properties.Resources.URRU);
            trayIconList.Add("URRG", Properties.Resources.URRG);
            trayIconList.Add("URRY", Properties.Resources.URRY);
            trayIconList.Add("URRR", Properties.Resources.URRR);
            trayIconList.Add("GUUU", Properties.Resources.GUUU);
            trayIconList.Add("GUUG", Properties.Resources.GUUG);
            trayIconList.Add("GUUY", Properties.Resources.GUUY);
            trayIconList.Add("GUUR", Properties.Resources.GUUR);
            trayIconList.Add("GUGU", Properties.Resources.GUGU);
            trayIconList.Add("GUGG", Properties.Resources.GUGG);
            trayIconList.Add("GUGY", Properties.Resources.GUGY);
            trayIconList.Add("GUGR", Properties.Resources.GUGR);
            trayIconList.Add("GUYU", Properties.Resources.GUYU);
            trayIconList.Add("GUYG", Properties.Resources.GUYG);
            trayIconList.Add("GUYY", Properties.Resources.GUYY);
            trayIconList.Add("GUYR", Properties.Resources.GUYR);
            trayIconList.Add("GURU", Properties.Resources.GURU);
            trayIconList.Add("GURG", Properties.Resources.GURG);
            trayIconList.Add("GURY", Properties.Resources.GURY);
            trayIconList.Add("GURR", Properties.Resources.GURR);
            trayIconList.Add("GGUU", Properties.Resources.GGUU);
            trayIconList.Add("GGUG", Properties.Resources.GGUG);
            trayIconList.Add("GGUY", Properties.Resources.GGUY);
            trayIconList.Add("GGUR", Properties.Resources.GGUR);
            trayIconList.Add("GGGU", Properties.Resources.GGGU);
            trayIconList.Add("GGGG", Properties.Resources.GGGG);
            trayIconList.Add("GGGY", Properties.Resources.GGGY);
            trayIconList.Add("GGGR", Properties.Resources.GGGR);
            trayIconList.Add("GGYU", Properties.Resources.GGYU);
            trayIconList.Add("GGYG", Properties.Resources.GGYG);
            trayIconList.Add("GGYY", Properties.Resources.GGYY);
            trayIconList.Add("GGYR", Properties.Resources.GGYR);
            trayIconList.Add("GGRU", Properties.Resources.GGRU);
            trayIconList.Add("GGRG", Properties.Resources.GGRG);
            trayIconList.Add("GGRY", Properties.Resources.GGRY);
            trayIconList.Add("GGRR", Properties.Resources.GGRR);
            trayIconList.Add("GYUU", Properties.Resources.GYUU);
            trayIconList.Add("GYUG", Properties.Resources.GYUG);
            trayIconList.Add("GYUY", Properties.Resources.GYUY);
            trayIconList.Add("GYUR", Properties.Resources.GYUR);
            trayIconList.Add("GYGU", Properties.Resources.GYGU);
            trayIconList.Add("GYGG", Properties.Resources.GYGG);
            trayIconList.Add("GYGY", Properties.Resources.GYGY);
            trayIconList.Add("GYGR", Properties.Resources.GYGR);
            trayIconList.Add("GYYU", Properties.Resources.GYYU);
            trayIconList.Add("GYYG", Properties.Resources.GYYG);
            trayIconList.Add("GYYY", Properties.Resources.GYYY);
            trayIconList.Add("GYYR", Properties.Resources.GYYR);
            trayIconList.Add("GYRU", Properties.Resources.GYRU);
            trayIconList.Add("GYRG", Properties.Resources.GYRG);
            trayIconList.Add("GYRY", Properties.Resources.GYRY);
            trayIconList.Add("GYRR", Properties.Resources.GYRR);
            trayIconList.Add("GRUU", Properties.Resources.GRUU);
            trayIconList.Add("GRUG", Properties.Resources.GRUG);
            trayIconList.Add("GRUY", Properties.Resources.GRUY);
            trayIconList.Add("GRUR", Properties.Resources.GRUR);
            trayIconList.Add("GRGU", Properties.Resources.GRGU);
            trayIconList.Add("GRGG", Properties.Resources.GRGG);
            trayIconList.Add("GRGY", Properties.Resources.GRGY);
            trayIconList.Add("GRGR", Properties.Resources.GRGR);
            trayIconList.Add("GRYU", Properties.Resources.GRYU);
            trayIconList.Add("GRYG", Properties.Resources.GRYG);
            trayIconList.Add("GRYY", Properties.Resources.GRYY);
            trayIconList.Add("GRYR", Properties.Resources.GRYR);
            trayIconList.Add("GRRU", Properties.Resources.GRRU);
            trayIconList.Add("GRRG", Properties.Resources.GRRG);
            trayIconList.Add("GRRY", Properties.Resources.GRRY);
            trayIconList.Add("GRRR", Properties.Resources.GRRR);
            trayIconList.Add("YUUU", Properties.Resources.YUUU);
            trayIconList.Add("YUUG", Properties.Resources.YUUG);
            trayIconList.Add("YUUY", Properties.Resources.YUUY);
            trayIconList.Add("YUUR", Properties.Resources.YUUR);
            trayIconList.Add("YUGU", Properties.Resources.YUGU);
            trayIconList.Add("YUGG", Properties.Resources.YUGG);
            trayIconList.Add("YUGY", Properties.Resources.YUGY);
            trayIconList.Add("YUGR", Properties.Resources.YUGR);
            trayIconList.Add("YUYU", Properties.Resources.YUYU);
            trayIconList.Add("YUYG", Properties.Resources.YUYG);
            trayIconList.Add("YUYY", Properties.Resources.YUYY);
            trayIconList.Add("YUYR", Properties.Resources.YUYR);
            trayIconList.Add("YURU", Properties.Resources.YURU);
            trayIconList.Add("YURG", Properties.Resources.YURG);
            trayIconList.Add("YURY", Properties.Resources.YURY);
            trayIconList.Add("YURR", Properties.Resources.YURR);
            trayIconList.Add("YGUU", Properties.Resources.YGUU);
            trayIconList.Add("YGUG", Properties.Resources.YGUG);
            trayIconList.Add("YGUY", Properties.Resources.YGUY);
            trayIconList.Add("YGUR", Properties.Resources.YGUR);
            trayIconList.Add("YGGU", Properties.Resources.YGGU);
            trayIconList.Add("YGGG", Properties.Resources.YGGG);
            trayIconList.Add("YGGY", Properties.Resources.YGGY);
            trayIconList.Add("YGGR", Properties.Resources.YGGR);
            trayIconList.Add("YGYU", Properties.Resources.YGYU);
            trayIconList.Add("YGYG", Properties.Resources.YGYG);
            trayIconList.Add("YGYY", Properties.Resources.YGYY);
            trayIconList.Add("YGYR", Properties.Resources.YGYR);
            trayIconList.Add("YGRU", Properties.Resources.YGRU);
            trayIconList.Add("YGRG", Properties.Resources.YGRG);
            trayIconList.Add("YGRY", Properties.Resources.YGRY);
            trayIconList.Add("YGRR", Properties.Resources.YGRR);
            trayIconList.Add("YYUU", Properties.Resources.YYUU);
            trayIconList.Add("YYUG", Properties.Resources.YYUG);
            trayIconList.Add("YYUY", Properties.Resources.YYUY);
            trayIconList.Add("YYUR", Properties.Resources.YYUR);
            trayIconList.Add("YYGU", Properties.Resources.YYGU);
            trayIconList.Add("YYGG", Properties.Resources.YYGG);
            trayIconList.Add("YYGY", Properties.Resources.YYGY);
            trayIconList.Add("YYGR", Properties.Resources.YYGR);
            trayIconList.Add("YYYU", Properties.Resources.YYYU);
            trayIconList.Add("YYYG", Properties.Resources.YYYG);
            trayIconList.Add("YYYY", Properties.Resources.YYYY);
            trayIconList.Add("YYYR", Properties.Resources.YYYR);
            trayIconList.Add("YYRU", Properties.Resources.YYRU);
            trayIconList.Add("YYRG", Properties.Resources.YYRG);
            trayIconList.Add("YYRY", Properties.Resources.YYRY);
            trayIconList.Add("YYRR", Properties.Resources.YYRR);
            trayIconList.Add("YRUU", Properties.Resources.YRUU);
            trayIconList.Add("YRUG", Properties.Resources.YRUG);
            trayIconList.Add("YRUY", Properties.Resources.YRUY);
            trayIconList.Add("YRUR", Properties.Resources.YRUR);
            trayIconList.Add("YRGU", Properties.Resources.YRGU);
            trayIconList.Add("YRGG", Properties.Resources.YRGG);
            trayIconList.Add("YRGY", Properties.Resources.YRGY);
            trayIconList.Add("YRGR", Properties.Resources.YRGR);
            trayIconList.Add("YRYU", Properties.Resources.YRYU);
            trayIconList.Add("YRYG", Properties.Resources.YRYG);
            trayIconList.Add("YRYY", Properties.Resources.YRYY);
            trayIconList.Add("YRYR", Properties.Resources.YRYR);
            trayIconList.Add("YRRU", Properties.Resources.YRRU);
            trayIconList.Add("YRRG", Properties.Resources.YRRG);
            trayIconList.Add("YRRY", Properties.Resources.YRRY);
            trayIconList.Add("YRRR", Properties.Resources.YRRR);
            trayIconList.Add("RUUU", Properties.Resources.RUUU);
            trayIconList.Add("RUUG", Properties.Resources.RUUG);
            trayIconList.Add("RUUY", Properties.Resources.RUUY);
            trayIconList.Add("RUUR", Properties.Resources.RUUR);
            trayIconList.Add("RUGU", Properties.Resources.RUGU);
            trayIconList.Add("RUGG", Properties.Resources.RUGG);
            trayIconList.Add("RUGY", Properties.Resources.RUGY);
            trayIconList.Add("RUGR", Properties.Resources.RUGR);
            trayIconList.Add("RUYU", Properties.Resources.RUYU);
            trayIconList.Add("RUYG", Properties.Resources.RUYG);
            trayIconList.Add("RUYY", Properties.Resources.RUYY);
            trayIconList.Add("RUYR", Properties.Resources.RUYR);
            trayIconList.Add("RURU", Properties.Resources.RURU);
            trayIconList.Add("RURG", Properties.Resources.RURG);
            trayIconList.Add("RURY", Properties.Resources.RURY);
            trayIconList.Add("RURR", Properties.Resources.RURR);
            trayIconList.Add("RGUU", Properties.Resources.RGUU);
            trayIconList.Add("RGUG", Properties.Resources.RGUG);
            trayIconList.Add("RGUY", Properties.Resources.RGUY);
            trayIconList.Add("RGUR", Properties.Resources.RGUR);
            trayIconList.Add("RGGU", Properties.Resources.RGGU);
            trayIconList.Add("RGGG", Properties.Resources.RGGG);
            trayIconList.Add("RGGY", Properties.Resources.RGGY);
            trayIconList.Add("RGGR", Properties.Resources.RGGR);
            trayIconList.Add("RGYU", Properties.Resources.RGYU);
            trayIconList.Add("RGYG", Properties.Resources.RGYG);
            trayIconList.Add("RGYY", Properties.Resources.RGYY);
            trayIconList.Add("RGYR", Properties.Resources.RGYR);
            trayIconList.Add("RGRU", Properties.Resources.RGRU);
            trayIconList.Add("RGRG", Properties.Resources.RGRG);
            trayIconList.Add("RGRY", Properties.Resources.RGRY);
            trayIconList.Add("RGRR", Properties.Resources.RGRR);
            trayIconList.Add("RYUU", Properties.Resources.RYUU);
            trayIconList.Add("RYUG", Properties.Resources.RYUG);
            trayIconList.Add("RYUY", Properties.Resources.RYUY);
            trayIconList.Add("RYUR", Properties.Resources.RYUR);
            trayIconList.Add("RYGU", Properties.Resources.RYGU);
            trayIconList.Add("RYGG", Properties.Resources.RYGG);
            trayIconList.Add("RYGY", Properties.Resources.RYGY);
            trayIconList.Add("RYGR", Properties.Resources.RYGR);
            trayIconList.Add("RYYU", Properties.Resources.RYYU);
            trayIconList.Add("RYYG", Properties.Resources.RYYG);
            trayIconList.Add("RYYY", Properties.Resources.RYYY);
            trayIconList.Add("RYYR", Properties.Resources.RYYR);
            trayIconList.Add("RYRU", Properties.Resources.RYRU);
            trayIconList.Add("RYRG", Properties.Resources.RYRG);
            trayIconList.Add("RYRY", Properties.Resources.RYRY);
            trayIconList.Add("RYRR", Properties.Resources.RYRR);
            trayIconList.Add("RRUU", Properties.Resources.RRUU);
            trayIconList.Add("RRUG", Properties.Resources.RRUG);
            trayIconList.Add("RRUY", Properties.Resources.RRUY);
            trayIconList.Add("RRUR", Properties.Resources.RRUR);
            trayIconList.Add("RRGU", Properties.Resources.RRGU);
            trayIconList.Add("RRGG", Properties.Resources.RRGG);
            trayIconList.Add("RRGY", Properties.Resources.RRGY);
            trayIconList.Add("RRGR", Properties.Resources.RRGR);
            trayIconList.Add("RRYU", Properties.Resources.RRYU);
            trayIconList.Add("RRYG", Properties.Resources.RRYG);
            trayIconList.Add("RRYY", Properties.Resources.RRYY);
            trayIconList.Add("RRYR", Properties.Resources.RRYR);
            trayIconList.Add("RRRU", Properties.Resources.RRRU);
            trayIconList.Add("RRRG", Properties.Resources.RRRG);
            trayIconList.Add("RRRY", Properties.Resources.RRRY);
            trayIconList.Add("RRRR", Properties.Resources.RRRR);
        }

        /// <summary>
        /// Set job's list tile size
        /// </summary>
        private void SetTileSize()
        {
            try
            {
                this.lvwJobs.TileSize = new Size(this.lvwJobs.Width - 22, this.lvwJobs.TileSize.Height);
            }
            catch
            {
                //In extreme sercumstances the SetTileSize can be triggered while lvwJobs is being drawn
                // This will cause the lvwJobs.Width to be 0 (zero), witch will cause an exception when
                // Tile size with is set to something negative
            }
        }

        /// <summary>
        /// Update forms GUI
        /// </summary>
        private void UpdateForm()
        {
            UpdateToolbar();
            UpdateJobList();
            UpdateTray();
        }

        /// <summary>
        /// Update main window application toolbar
        /// </summary>
        private void UpdateToolbar()
        {
            tsbtnRun.Checked = bsmApp.IsRunning;
            tsbtnRun.Text = (tsbtnRun.Checked) ? Properties.Resources.SchedulerOn : Properties.Resources.SchedulerOff;
        }

        /// <summary>
        /// Update listview with jobs and their stats.
        /// </summary>
        private void UpdateJobList()
        {
            //By updating tiles instead of removing all and adding them anew
            // we can keep the user selected items.
            // If itemes gets unselected, contect menu doesn't work
            // and label edit is canceled.

            string[] jobTitles = bsmApp.JobTitles;

            lvwJobs.SuspendLayout();
            string[] selectedItemsText = GetSelectedItemsText();

            foreach (string jobTitle in jobTitles)
            {
                if (lvwJobs.Items.ContainsKey(jobTitle))
                {
                    UpdateJobTile(jobTitle);
                }
                else
                {
                    AddJobTile(jobTitle);
                }
            }
            DeleteOldJobTiles(jobTitles);
            ReSelectItems(selectedItemsText);

            //Prevent label edit from canceling on update
            if (!LabelBeingEdited())
                lvwJobs.Sort();
            
            lvwJobs.PerformLayout();
        }

        private string[] GetSelectedItemsText()
        {
            string[] list;
            ListViewItem[] selectedItems = new ListViewItem[lvwJobs.SelectedItems.Count];
            list = new string[selectedItems.Length];

            lvwJobs.SelectedItems.CopyTo(selectedItems, 0);
            int i = 0;
            foreach (ListViewItem item in selectedItems)
            {
                list[i] = item.Text;
                i++;
            }
            return list;
        }

        private void ReSelectItems(string[] itemKeys)
        {
            foreach (string key in itemKeys)
            {
                if (lvwJobs.Items.ContainsKey(key))
                {
                    lvwJobs.Items[key].Selected = true;
                }
            }
        }

        private bool LabelBeingEdited()
        {
            return labelEdited != String.Empty;
        }

        private bool LabelBeingEdited(string title)
        {
            return (labelEdited == title);
        }

        /// <summary>
        /// Update a listitem/tile in job list view
        /// </summary>
        /// <param name="jobTitle">job title</param>
        private void UpdateJobTile(string jobTitle)
        {
            string jobTypeName = bsmApp.GetJobTypeName(jobTitle);
            ListViewItem item = lvwJobs.Items[jobTitle];
            SetJobTile(jobTitle, jobTypeName, item);
        }

        /// <summary>
        /// Add a listitem/tile to jobslist
        /// </summary>
        /// <param name="jobTitle">job title</param>
        private void AddJobTile(string jobTitle)
        {
            string jobTypeName = bsmApp.GetJobTypeName(jobTitle);
           ListViewItem item = lvwJobs.Items.Add(jobTitle, jobTitle, jobTypeName);
           item.SubItems.Add("");
           item.SubItems.Add("");
           SetJobTile(jobTitle, jobTypeName, item);
        }

        /// <summary>
        /// Set the status for a tile in the job list view
        /// </summary>
        /// <param name="jobTitle">job title</param>
        /// <param name="jobTypeName">job type </param>
        /// <param name="item">list view item, the tile being set</param>
        private void SetJobTile(string jobTitle, string jobTypeName, ListViewItem item)
        {
            Status status = bsmApp.GetJobStatus(jobTitle);

            string typeDate = String.Format("{0}  {1}",
                jobTypeName,
                status.Date.ToString());

            string description = status.Text;
            
            // Prevents label edits to be cancelled due to updates of the listview.
            if (!LabelBeingEdited(jobTitle))
            {
                item.Text = jobTitle;
                item.SubItems[1].Text = typeDate;
                item.SubItems[2].Text = description;
            }

            item.StateImageIndex = (int)status.Code;
            if (bsmApp.DisplayJobInTray(jobTitle))
            {
                item.Group = lvwJobs.Groups[1];
            }
            else
            {
                item.Group = lvwJobs.Groups[0];
            }

            item.BackColor = bsmApp.GetJobEnabled(jobTitle) ?
                lvwJobs.BackColor :
                Color.LightGray;
        }

        /// <summary>
        /// Delete job tiles, list view items, no longer in use
        /// </summary>
        /// <param name="jobTitles">list of current jobs, given by job titles</param>
        private void DeleteOldJobTiles(string[] jobTitles)
        {
            // An item has been deleted from scheduler
            // and should be deleted from listview
            int dif = lvwJobs.Items.Count - jobTitles.Length;
            if (dif != 0)
            {
                ListViewItem[] items = new ListViewItem[lvwJobs.Items.Count];
                lvwJobs.Items.CopyTo(items, 0);
                foreach (ListViewItem item in items)
                {
                    if (!bsmApp.GetJobExists(item.Text))
                    {
                        lvwJobs.Items.RemoveByKey(item.Text);
                        dif--;
                        if (dif == 0)
                            break;
                    }
                }
            }
        }

        /// <summary>
        /// Update the tray application and its menu
        /// </summary>
        private void UpdateTray()
        {
            // without the smooth update alternative 
            // the context menu flickers if it gets updates while it is displayed
            bool smoothUpdate = cmsApplication.Visible;

            if (!smoothUpdate)
            {
                foreach (ToolStripMenuItem item in jobsOnTrayMenu)
                {
                    cmsApplication.Items.Remove(item);
                    item.Dispose();
                }
                jobsOnTrayMenu.Clear();
            }

            string[] jobsInTray = bsmApp.JobsInTray;

            if (jobsInTray.Length == 0)
            {
                UpdateTrayForNoJobsInTray();
            }
            else
            {
                UpdateTrayJobsInTray(jobsInTray, smoothUpdate);
            }

        }

        /// <summary>
        /// Update method for when at least one job is displayed in tray application
        /// </summary>
        /// <param name="jobsInTray">array of jobtitles for jobs displayed in tray application</param>
        private void UpdateTrayJobsInTray(string[] jobsInTray, bool smoothUpdate)
        {
            mnuApplicationSep.Visible = true;
            int menuIndex = 0;
            string iconKey = "";

            foreach (string title in jobsInTray)
            {
                if (smoothUpdate)
                {
                    UpdateJobOnTray(menuIndex, title, ref iconKey);
                }
                else
                {
                    AddJobToTray(menuIndex, title, ref iconKey);
                }
                menuIndex++;
            }

            if (jobsInTray.Length > 1)
            {
                iconKey = iconKey.PadRight(4, 'U');
            }
            

            SetTrayIcon(iconKey);
        }

        /// <summary>
        /// Set tray application icon
        /// </summary>
        /// <param name="iconKey">a key matching an icon in the trayIconList</param>
        private void SetTrayIcon(string iconKey)
        {
            if (trayIconList.ContainsKey(iconKey))
            {
                tray.Icon = trayIconList[iconKey];
            }
            else
            {
                tray.Icon = trayIconList["App"];
            }
        }

        /// <summary>
        /// Adds a job, based on its title.
        /// </summary>
        /// <param name="menuIndex">index for where on the menu the job should be placed</param>
        /// <param name="jobTitle">title of the job to be added</param>
        /// <param name="iconKey">ref string accumulates statuses codes for jobs so the right icon can be displayed</param>
        private void AddJobToTray(int menuIndex, string jobTitle, ref string iconKey)
        {
            try
            {
                Status status = bsmApp.GetJobStatus(jobTitle);
                
                ToolStripMenuItem item = new ToolStripMenuItem();
                item.Click += this.mnuApplicationJobItem_Click;
                item.Tag = jobTitle;

                SetJobOnTrayMenu(item, jobTitle, status, ref iconKey);

                cmsApplication.Items.Insert(menuIndex, item);
                jobsOnTrayMenu.Add(item);
            }
            catch (BsmException ex)
            {
                bsmApp.WriteExceptionToLog(ex);
                bsmApp.DisplayMessage(new UIException(ex.Message, ex));
            }
        }

        /// <summary>
        /// Prepares a menu item för displaying a job on the tray menu.
        /// </summary>
        /// <param name="item">mneu item</param>
        /// <param name="jobTitle">job title</param>
        /// <param name="status">job status</param>
        /// <param name="iconKey">ref string accumulates statuses codes for jobs so the right icon can be displayed</param>
        private void SetJobOnTrayMenu(ToolStripMenuItem item, string jobTitle, Status status, ref string iconKey)
        {

            item.Text = String.Format("{0} {1}", jobTitle, status.Date.ToString());
            item.Image = this.ilsStateIcons.Images[status.Code.ToString()];

            if (!bsmApp.GetJobEnabled(jobTitle))
            {
                Font font = new Font(item.Font, FontStyle.Italic);
                item.Font = font;
            }
            else
            {
                Font font = new Font(item.Font, FontStyle.Regular);
                item.Font = font;
            }

            iconKey += status.Code.ToString().Substring(0, 1);
        }

        /// <summary>
        /// Updates jobs on tray, based on its title.
        /// </summary>
        /// <param name="menuIndex">index for where on the menu the job should be placed</param>
        /// <param name="jobTitle">title of the job to be added</param>
        /// <param name="iconKey">ref string accumulates statuses codes for jobs so the right icon can be displayed</param>
        private void UpdateJobOnTray(int menuIndex, string jobTitle, ref string iconKey)
        {
            try
            {
                Status status = bsmApp.GetJobStatus(jobTitle);
                ToolStripMenuItem item = (ToolStripMenuItem)cmsApplication.Items[menuIndex];
                SetJobOnTrayMenu(item, jobTitle, status, ref iconKey);
            }
            catch (BsmException ex)
            {
                bsmApp.WriteExceptionToLog(ex);
                bsmApp.DisplayMessage(new UIException(ex.Message, ex));
            }
        }

        /// <summary>
        /// Update method for tray application menu when no jobs are displayed
        /// </summary>
        private void UpdateTrayForNoJobsInTray()
        {
            mnuApplicationSep.Visible = false;
            tray.Icon = trayIconList["App"];
        }

        #endregion InitAndUpdate

        #region PublicMethods

        public void DisplayMessage(string message, string title)
        {
            try
            {
                if (tray != null)
                {
                    if (message != String.Empty || title != String.Empty)
                    {
                        tray.BalloonTipText = message;
                        tray.BalloonTipTitle = title;
                        tray.ShowBalloonTip(10);
                    }
                }
            }
            catch
            {
                Console.WriteLine("Error Display Message");
            }
        }

        #endregion PublicMethods

        #region ToolbarEventHandlers

        private void tsbtnNew_Click(object sender, EventArgs e)
        {
            if (bsmApp.CreateJob(this))
            {
                UpdateForm();
            }
        }

        private void tsbtnDelete_Click(object sender, EventArgs e)
        {
            foreach (ListViewItem item in lvwJobs.SelectedItems)
            {
                bsmApp.RemoveJob(item.Text);
            }
            if (lvwJobs.SelectedItems.Count > 0)
                UpdateForm();
        }

        private void tsbtnRun_Click(object sender, EventArgs e)
        {
            if (tsbtnRun.Checked)
            {
                bsmApp.AskStartScheduler();
            }
            else
            {
                bsmApp.AskStopScheduler();
            }
        }

        private void tsbtnRefresh_Click(object sender, EventArgs e)
        {
            bsmApp.AskRefresh();
        }

        private void tsbtnConfigApplication_Click(object sender, EventArgs e)
        {
            bsmApp.ConfigureApplication(this);
        }

        #endregion ToolbarEventHandlers

        #region JobListEventHandlers

        /// <summary>
        /// Eventhandler Key down on job listview
        /// Checks if delete key was pressed
        /// </summary>
        private void lvwJobs_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Delete)
            {
                foreach (ListViewItem item in lvwJobs.SelectedItems)
                {
                    bsmApp.RemoveJob(item.Text);
                }
                UpdateForm();
            }
            else
            {
                MainFrame_StartDebugMode(sender, e);
            }
        }

        /// <summary>
        /// Event handler: job list view double click
        /// </summary>
        private void lvwJobs_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            ListViewItem item = lvwJobs.GetItemAt(e.X, e.Y);
            if (item != null)
            {
                switch (bsmApp.MainFormActionOnClick)
                {
                    case ActionOnClick.None:
                        //Do nothing
                        break;
                    case ActionOnClick.Configure:
                        bsmApp.ConfigureJob(item.Text, this);
                        break;
                    case ActionOnClick.Launch:
                        bsmApp.LaunchJobApplication(item.Text);
                        break;
                }
            }
        }

        /// <summary>
        /// Event handler: rezise jobs list view
        /// Set tile siez to match the new size
        /// /// </summary>
        private void lvwJobs_Resize(object sender, System.EventArgs e)
        {
            SetTileSize();
        }

        /// <summary>
        /// Event handler: Job list view mouse down. Beginning of drag and drop
        /// </summary>
        private void lvwJobs_MouseDown(object sender, MouseEventArgs e)
        {
            if ((e.Button == MouseButtons.Left) && (e.Clicks == 1))
            {
                mouseDown = true;
            }
            else
            {
                mouseDown = false;
            }
        }

        /// <summary>
        /// Event handler: Job List view mouse move. Activates drag and drop
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void lvwJobs_MouseMove(object sender, MouseEventArgs e)
        {
            try
            {
                if (mouseDown && (e.Button == MouseButtons.Left))
                {
                    string[] titles = new string[lvwJobs.SelectedItems.Count];
                    int i = 0;
                    foreach (ListViewItem item in lvwJobs.SelectedItems)
                    {
                        titles[i] = item.Text;
                        i++;
                    }
                    lvwJobs.DoDragDrop(String.Join(Environment.NewLine, titles), DragDropEffects.All);
                }
            }
            catch
            {
                Console.WriteLine();
                Console.WriteLine("Error mouse move");
                Console.WriteLine();
                mouseDown = false;
            }
        }

        /// <summary>
        /// Event handler: Job list view, Drag enter event
        /// Only accept drag and drop within the listview
        /// </summary>
        private void lvwJobs_DragEnter(object sender, DragEventArgs e)
        {
            try
            {
                if (e.Data.GetDataPresent(DataFormats.Text) && mouseDown)
                {
                    e.Effect = DragDropEffects.Move;
                }
                else
                {
                    mouseDown = false;
                    e.Effect = DragDropEffects.None;
                }
            }
            catch
            {
                Console.WriteLine();
                Console.WriteLine("Error DragEnter");
                Console.WriteLine();
                mouseDown = false;
            }
        }

        /// <summary>
        /// Event handler: Job list view, Drag and Drop event
        /// Only allow drag and drop within listview
        /// </summary>
        private void lvwJobs_DragDrop(object sender, DragEventArgs e)
        {
            try
            {
                if (mouseDown && (lvwJobs.Equals(sender)))
                {
                    mouseDown = false;
                    bool targetAreaDisplayedInTray = GetTargetAreaDisplayInTray(e);
                    string[] jobTitles = GetTitlesFromSource(e);
                    foreach (string title in jobTitles)
                    {

                        SetDisplayInTrayPropertyForJob(title, targetAreaDisplayedInTray);
                    }
                    if (jobTitles.Length > 0)
                        UpdateForm();
                }
            }
            catch
            {
                Console.WriteLine();
                Console.WriteLine("Error DragDrop");
                Console.WriteLine();
                mouseDown = false;
            }
        }

        /// <summary>
        /// Get string-array if tile texts from the source of a drag event.
        /// </summary>
        /// <param name="e">Drag event arguments</param>
        /// <returns>string[] : list of selected job titles in job list view</returns>
        private string[] GetTitlesFromSource(DragEventArgs e)
        {
            string source = e.Data.GetData(DataFormats.Text).ToString();
            string[] jobTitles = source.Split(Environment.NewLine.ToCharArray(), StringSplitOptions.RemoveEmptyEntries);
            return jobTitles;
        }

        /// <summary>
        /// Returns true or false depending on the drop area.
        /// True if the drop area is that for jobs displayed in the tray.
        /// </summary>
        /// <param name="e">Drag event arguments</param>
        /// <returns>bool : true if the drop area is display job in tray</returns>
        private bool GetTargetAreaDisplayInTray(DragEventArgs e)
        {
            Point localPoint = lvwJobs.PointToClient(new Point(e.X, e.Y));
            ListViewItem item = lvwJobs.GetItemAt(localPoint.X, localPoint.Y);
            bool targetAreaDisplayedInTray;
            if (item != null)
            {
                targetAreaDisplayedInTray = bsmApp.DisplayJobInTray(item.Text);
            }
            else
            {
                targetAreaDisplayedInTray = true;
            }
            return targetAreaDisplayedInTray;
        }

        /// <summary>
        /// Set the display in tray property for a job by its title
        /// </summary>
        /// <param name="title">job title</param>
        /// <param name="displayInTray">bool : new value</param>
        private void SetDisplayInTrayPropertyForJob(string title, bool displayInTray)
        {
            bool sourceItemDisplayedInTray = bsmApp.DisplayJobInTray(title);
            if (displayInTray != sourceItemDisplayedInTray)
            {
                if (displayInTray)
                {
                    //Move to being displayed in tray application menu
                    //if there are spots stil available
                    if (bsmApp.TraySpotsAvailable())
                    {
                        bsmApp.AddJobToTray(title);
                    }
                }
                else
                {
                    //Move from being display to not being display
                    //Always safe
                    bsmApp.RemoveJobFromTray(title);
                }
            }
        }

        /// <summary>
        /// Store the label being edited so update of that tile can be prevented.
        /// Otherwise the label edit gets cancelled on update.
        /// </summary>
        private void lvwJobs_BeforeLabelEdit(object sender, LabelEditEventArgs e)
        {
            labelEdited = e.Label;
        }

        /// <summary>
        /// Eventhandler - Edit job title
        /// </summary>
        private void lvwJobs_AfterLabelEdit(object sender, LabelEditEventArgs e)
        {
            labelEdited = String.Empty;
            if (e.Label == null)
            {
                e.CancelEdit = true;
                return;
            }
            string currentJobTitle = lvwJobs.Items[e.Item].Text;
            if (bsmApp.SetNewJobTitle(currentJobTitle, e.Label))
                UpdateForm();
            // Always update since label edit prevents the selected item from being updated.
            

            // Always cancel edit!!
            // Since we redraw the listview while in the event
            // The list view will be updated after we exit, hence change our
            // already correctly drawn list view.
            e.CancelEdit = true;
        }


        #endregion JobListEventHandlers

        #region MainFrameEventHandlers

        /// <summary>
        /// Event handler: Main Form closing
        /// Prevents it from closing if it is user induced.
        /// </summary>
        private void MainFrame_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (e.CloseReason == CloseReason.UserClosing)
            {
                e.Cancel = true;
                this.Hide();
            }
            
        }

        #endregion MainFrameEventHandlers

        #region JobContextMenuEventHandlers        


        /// <summary>
        /// Job context menu: Configure a job
        /// </summary>
        private void mnuJobConfigure_Click(object sender, EventArgs e)
        {
            if (lvwJobs.SelectedItems.Count > 0)
            {
                string jobTitle = lvwJobs.SelectedItems[0].Text;
                bsmApp.ConfigureJob(jobTitle, this);
            }
        }

        /// <summary>
        /// Job context menu: Launch Monitored Application
        /// </summary>
        private void mnuJobLaunch_Click(object sender, EventArgs e)
        {
            if (lvwJobs.SelectedItems.Count > 0)
            {
                string jobTitle = lvwJobs.SelectedItems[0].Text;
                bsmApp.LaunchJobApplication(jobTitle);
            }
        }

        /// <summary>
        /// Job context menu: Set display in tray
        /// </summary>
        private void mnuJobDisplayInTray_Click(object sender, EventArgs e)
        {
            foreach (ListViewItem item in lvwJobs.SelectedItems)
            {
                if (mnuJobDisplayInTray.Checked)
                {
                    bsmApp.AddJobToTray(item.Text);
                }
                else
                {
                    bsmApp.RemoveJobFromTray(item.Text);
                }
            }
            UpdateForm();
        }

        /// <summary>
        /// Job context menu: Set run interval
        /// </summary>
        private void mnuJobRunInterval_Click(object sender, EventArgs e)
        {
            if (lvwJobs.SelectedItems.Count > 0)
            {
                IDataField<int> runInterval = new DataField<int>(new DataBinding.Validators.LargerThanZeroValidator<int>(),
                    null, 0);
                ListViewItem[] selectedjobs = new ListViewItem[lvwJobs.SelectedItems.Count];
                lvwJobs.SelectedItems.CopyTo(selectedjobs, 0);

                runInterval.Value = bsmApp.GetJobRunInterval(lvwJobs.SelectedItems[0].Text);
                if (InputBox<int>.Show(runInterval, 
                    Properties.Resources.InvalidInputJobRunInterval,
                    Properties.Resources.InputBoxRunIntervalCaption,
                    Properties.Resources.InputBoxRunIntervalPrompt) == DialogResult.OK)
                {
                    foreach (ListViewItem item in selectedjobs)
                    {
                        bsmApp.SetJobRunInterval(item.Text, runInterval.Value);
                    }
                }
            }
        }

        /// <summary>
        /// Event Handler: Job context menu opens
        /// Updates the state of the menu items to match the options
        /// available depending on item selections.
        /// </summary>
        private void cmsJobManagement_Opening(object sender, CancelEventArgs e)
        {
            if (lvwJobs.SelectedItems.Count > 0)
            {
                cmsJobManagement.Enabled = true;
                SetConfigMenuEnableProperty();
                SetLaunchMenuEnableProperty();
                SetDisplayInTrayMenuEnableProperty();
                SetEnabledEnableProperty();
            }
            else
            {
                mnuJobDisplayInTray.Checked = false;
                cmsJobManagement.Enabled = false;
            }
        }

        /// <summary>
        /// Set how the Configure option should be displayed on the context menu
        /// </summary>
        private void SetConfigMenuEnableProperty()
        {
            if (lvwJobs.SelectedItems.Count > 1)
            {
                mnuJobConfigure.Enabled = false;
            }
            else
            {
                mnuJobConfigure.Enabled = true;
            }
        }

        /// <summary>
        /// Set if Launch Monitored application should be enabled
        /// </summary>
        private void SetLaunchMenuEnableProperty()
        {
            if (lvwJobs.SelectedItems.Count > 1)
            {
                mnuJobLaunch.Enabled = false;
            }
            else
            {
                mnuJobLaunch.Enabled = true;
            }
        }

        /// <summary>
        /// Set how the display in tray menu option should be displayed
        /// </summary>
        private void SetDisplayInTrayMenuEnableProperty()
        {
            bool displayedInTray = bsmApp.DisplayJobInTray(lvwJobs.SelectedItems[0].Text);
            mnuJobDisplayInTray.Checked = displayedInTray;
            for (int i = 1; i < lvwJobs.SelectedItems.Count; i++)
            {
                if (displayedInTray != bsmApp.DisplayJobInTray(lvwJobs.SelectedItems[i].Text))
                {
                    mnuJobDisplayInTray.CheckState = CheckState.Indeterminate;
                    break;
                }
            }
        }

        /// <summary>
        /// Set how the Job Enable property should be marked on the context menu
        /// </summary>
        private void SetEnabledEnableProperty()
        {
            bool enabled = bsmApp.GetJobEnabled(lvwJobs.SelectedItems[0].Text);
            mnuJobEnabled.Checked = enabled;
            for (int i = 1; i < lvwJobs.SelectedItems.Count; i++)
            {
                if (enabled != bsmApp.GetJobEnabled(lvwJobs.SelectedItems[i].Text))
                {
                    mnuJobEnabled.CheckState = CheckState.Indeterminate;
                    break;
                }
            }
        }

        /// <summary>
        /// Context menu: Job Enable property
        /// </summary>
        private void mnuJobEnabled_Click(object sender, EventArgs e)
        {
            foreach (ListViewItem item in lvwJobs.SelectedItems)
            {
                bsmApp.SetJobEnabledProperty(item.Text, mnuJobEnabled.Checked);
            }
            UpdateForm();
        }

        #endregion JobContextMenuEventHandlers

        #region TrayAndApplicationMenuEventHandlers

        private void mnuApplicationOpen_Click(object sender, EventArgs e)
        {
            ShowMainForm();
        }

        /// <summary>
        /// Show Main application form and do needed updates.
        /// </summary>
        private void ShowMainForm()
        {
            this.Show();
            UpdateForm();
            this.Focus();
        }

        private void tray_DoubleClick(object sender, EventArgs e)
        {
            ShowMainForm();
        }

        private void mnuApplicationExit_Click(object sender, EventArgs e)
        {
            ClosingApplication();
        }

        /// <summary>
        /// Closes the Application
        /// Should be called instead of this.Close() if the intention
        /// really is to exit the application.
        /// </summary>
        private void ClosingApplication()
        {
            try
            {
                threadSyncher.CancelAsync();
                bsmApp.SaveApplicationSettings();
                bsmApp.Close();
            }
            catch
            {
                //Do nothing
            }
            finally
            {
                System.Windows.Forms.Application.Exit();
            }
        }

        private void mnuApplicationAbout_Click(object sender, EventArgs e)
        {
            AboutBSM form = new AboutBSM();
            form.ShowDialog();
        }

        private void mnuApplicationConfigure_Click(object sender, EventArgs e)
        {
            bsmApp.ConfigureApplication(null);
        }

        private void mnuApplicationRefresh_Click(object sender, EventArgs e)
        {
            bsmApp.AskRefresh();
        }

        private void mnuApplicationJobItem_Click(object sender, EventArgs e)
        {
            ToolStripMenuItem item = sender as ToolStripMenuItem;
            if (item != null)
            {
                switch (bsmApp.TrayActionOnClick)
                {
                    case ActionOnClick.None:
                        //Do nothing
                        break;
                    case ActionOnClick.Configure:
                        bsmApp.ConfigureJob(item.Tag.ToString(), null);
                        break;
                    case ActionOnClick.Launch:
                        bsmApp.LaunchJobApplication(item.Tag.ToString());
                        break;
                }
            }
        }

        #endregion TrayAndApplicationMenuEventHandlers

        #region ThreadSyncherEventHandlers

        /// <summary>
        /// Event handler: A call back has been issued to the BackgroundWorker/thread syncher.
        /// </summary>
        void threadSyncher_EventRaized(object sender, ProgressChangedEventArgs e)
        {
            BsmMessage message = e.UserState as BsmMessage;
            if (message != null)
            {
                switch (message.MessageType)
                {
                    case BsmMessageTypes.Refresh:
                    case BsmMessageTypes.StartingScheduler:
                    case BsmMessageTypes.StoppingScheduler:
                        UpdateToolbar();
                        break;
                    case BsmMessageTypes.JobStatusChanged:
                    case BsmMessageTypes.Exception:
                    case BsmMessageTypes.General:
                    case BsmMessageTypes.JobFinished:
                        UpdateForm();
                        break;
                    case BsmMessageTypes.UIException:
                        //Do not update form
                        //May cause infinite loop of messages
                        break;
                }
                if (message.Display)
                {
                    DisplayMessage(message.Text, message.Title);
                }
            }
        }

        #endregion ThreadSyncher

        #region Debug

        private void MainFrame_StartDebugMode(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.F10)
            {
                bsmApp.WriteDebugReport();
                
            }
        }

        #endregion //Debug



    }
}