//**********************************//
// Written by: Per Bhlin
// perbo770@student.liu.se
// Copyright: (c) 2009
//
// Implementation started: July 27, 2009
// Last modified: August 12, 2009
//**********************************//

using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using BSM.core;

namespace BSM
{

    /// <summary>
    /// Schedules jobs to be run at given intervals.
    /// Each job is running in its own thread.
    /// A job-run is invoked by timer, or direct call to RunJobs().
    /// If the scheduler is running when RunJobs() is invoked,
    /// all jobs are rescheduled.
    /// If the scheduler is running when a job is added, it runs
    /// immediately.
    /// RunJob(IJob) is called asynchronius by other threads.
    /// Because of this, all methods used within RunJob has been 
    /// made thread safe by using locks.
    /// 
    /// Usage:
    ///     See IScheduler
    /// 
    /// Throws:
    ///     JobDoesNotExistException
    ///     
    /// Events:
    ///     JobFinished
    ///     StatusChanged
    /// </summary>
    class Scheduler : BSM.IScheduler
    {
        IDictionary<string, IJob> jobs;
        IDictionary<string, Status> statuses;
        IDictionary<string, Timer> timers;
        //IDictionary<string, Thread> workerThreads;
        IDictionary<string, bool> workerThreads;

        bool isRunning;

        /// <summary>
        /// Default constructor
        /// </summary>
        public Scheduler()
        {
            jobs = new SortedList<string, IJob>();
            statuses = new Dictionary<string, Status>();
            timers = new Dictionary<string, Timer>();
            //workerThreads = new Dictionary<string, Thread>();
            workerThreads = new Dictionary<string, bool>();
            isRunning = false;
        }

        #region Events
        public event JobFinishedEventHandler JobFinished;

        /// <summary>
        /// Called to trigger the JobFinished event.
        /// </summary>
        /// <param name="args">arguments passed to the event handler</param>
        private void OnJobFinished(JobFinishedEventArgs args)
        {
            if (JobFinished != null)
            {
                JobFinished(this, args);
            }
        }

        public event StatusChangedEventHandler StatusChanged;

        /// <summary>
        /// Called to trigger the StatusChanged event.
        /// </summary>
        /// <param name="args">arguments passed to the event handler</param>
        private void OnStatusChanged(StatusChangedEventArgs args)
        {
            if (StatusChanged != null)
            {
                StatusChanged(this, args);
            }
        }
        #endregion //Events

        /// <summary>
        /// Add job to the scheduler
        /// </summary>
        /// <param name="job">job being added</param>
        public void AddJob(IJob job)
        {
            jobs.Add(job.Title, job);
            AddJobStatus(job, new Status()); //Thread safe
            if (job.Enabled)
                AddJobTimer(job);
        }

        /// <summary>
        /// Add job status entry
        /// Thread safe
        /// </summary>
        /// <param name="job">job</param>
        /// <param name="jobStatus">job status</param>
        private void AddJobStatus(IJob job, Status jobStatus)
        {
            lock (statuses)
            {
                statuses.Add(job.Title, jobStatus);
            }
        }

        /// <summary>
        /// Add timer for a job
        /// The timer will fire right away and then att given intervals
        /// depending on the jobs property RunIntervalInSeconds
        /// Timmer will only be set if scheduler is running.
        /// </summary>
        /// <param name="job">job</param>
        private void AddJobTimer(IJob job)
        {
            if (IsRunning)
            {
                timers.Add(job.Title, new Timer(RunJob, job, 0, job.RunIntervalInSeconds * 1000));
            }
        }

        /// <summary>
        /// Remove a job from the scheduler.
        /// </summary>
        /// <param name="title">the title of the job being removed</param>
        public void RemoveJob(string title)
        {
            try
            {
                IJob job = GetJob(title);
                lock (job)
                {
                    RemoveTimer(job);
                    RemoveJobStatus(job); //Thread safe
                    jobs.Remove(title);
                }
            }
            catch (KeyNotFoundException ex)
            {
                throw new JobDoesNotExistException(title, ex);
            }
        }

        /// <summary>
        /// Remove job's status object from list
        /// </summary>
        /// <param name="job">job</param>
        private void RemoveJobStatus(IJob job)
        {
            string key = job.Title;
            lock (statuses)
            {
                if (statuses.ContainsKey(key))
                {
                    statuses.Remove(key);
                }
            }
        }

        /// <summary>
        /// Remove the jobs timer from list of timers
        /// </summary>
        /// <param name="job">job</param>
        private void RemoveTimer(IJob job)
        {
            string key = job.Title;
            if (timers.ContainsKey(key))
            {
                timers[key].Dispose();
                timers.Remove(key);
            }
        }

        /// <summary>
        /// Checks whether a job is scheduled or not.
        /// </summary>
        /// <param name="title">the title of the job sought</param>
        /// <returns>true if the job is scheduled</returns>
        public bool GetJobExists(string title)
        {
            return jobs.ContainsKey(title);
        }

        /// <summary>
        /// Get job with a given title.
        /// </summary>
        /// <param name="title">the title of the job sought</param>
        /// <returns>job if it exists. otherwise it throws exception</returns>
        public IJob GetJob(string title)
        {
            return jobs[title];
        }

        /// <summary>
        /// Runs all jobs in scheduler.
        /// This method is not affected by if the scheduler is running or not.
        /// </summary>
        public void RunJobs()
        {
            if (isRunning)
            {
                UpdateTimers();
            }
            else
            {
               RunJobsOneTime();
            }
        }

        /// <summary>
        /// Runs all jobs one time, not using timers.
        /// </summary>
        private void RunJobsOneTime()
        {
            foreach (IJob job in jobs.Values)
            {
                if (job.Enabled)
                    RunJobInNewThread(job);
            }
        }

        private void RunJobInNewThread(IJob job)
        {
            Thread jobThread;
            jobThread = new Thread(new ParameterizedThreadStart(this.RunJob));
            jobThread.Start(job);
            Thread.Sleep(0);
        }

        /// <summary>
        /// Resset the job timmers so new jobs get scheduled right away.
        /// </summary>
        private void UpdateTimers()
        {
            foreach (IJob job in jobs.Values)
            {
                if (job.Enabled)
                    UpdateTimer(job);
            }
        }

        private void UpdateTimer(IJob job)
        {
            if (timers.ContainsKey(job.Title))
            {
                timers[job.Title].Change(0, job.RunIntervalInSeconds * 1000);
            }
            else
            {
                AddJobTimer(job);
            }
        }

        /// <summary>
        /// Wrapper function to match thread delegates
        /// Should not otherwised be used. Instead call RunJob(IJob)
        /// </summary>
        /// <param name="job"></param>
        private void RunJob(object job)
        {
            IJob properJob = job as IJob;
            if (properJob != null)
            {
                RunJob(properJob);
            }
        }

        /// <summary>
        /// Runs a job.
        /// Expected to run in its own thread
        /// If the job is already running, a new run will not be started.
        /// Thread entry point
        /// Thread-safe
        /// </summary>
        /// <param name="job">the job being run</param>
        private void RunJob(IJob job)
        {
            try
            {

                Status newStatus = null;
                try
                {
                    // place these operations in lock statement
                    // to make them atomic thanks to use of the same lock elsewhere.
                    lock (job)
                    {
                        //If job is already running, don't run
                        //This to prevent them from stacking up if update interval
                        //is set to less than the time it takes to run a job.
                        if (JobIsRunning(job) || !ThreadSafeJobExist(job))
                            return;

                        RegisterJobAsRunning(job, Thread.CurrentThread);
                    }

                    newStatus = new Status(job.Run());

                    Status oldStatus = GetJobStatus(job.Title);

                    // place these operations in lock statement
                    // to make them atomic thanks to use of the same lock elsewhere.
                    lock (job)
                    {
                        if (!ThreadSafeJobExist(job))
                            return;

                        SetJobStatus(job, newStatus);

                        // Invoke StatusChanged Event if needed
                        if (StatusHasChanged(newStatus, oldStatus))
                        {
                            OnStatusChanged(new StatusChangedEventArgs(job.Title));
                        }
                    }
                }
                catch (JobDoesNotExistException)
                {
                    //Job was removed while it was executing.
                    //Catch the error and do nothing
                }
                catch (RunFailedException ex)
                {
                    SetJobStatus(job, new Status(Status.StatusCode.Unknown, ex.Message));
                }
                catch
                {
                    SetJobStatus(job, new Status(Status.StatusCode.Unknown,
                        Properties.Resources.NullStatusText));
                }
                finally
                {
                    UnregisterJobAsRunning(job);
                    //Invoke JobFinished Event
                    OnJobFinished(new JobFinishedEventArgs(job.Title));
                }
            }
            catch
            {
                //Do nothing
                //Since this is an entry point for a thread we want to catch any exception that might happen
                //so it doesn't cause the entire application to go down.
            }
        }

        /// <summary>
        /// Returns true if the job exist in the scheduler
        /// Thread safe
        /// </summary>
        /// <param name="job">job</param>
        /// <returns>true if job exist in scheduler</returns>
        private bool ThreadSafeJobExist(IJob job)
        {
            lock (statuses)
            {
                return statuses.ContainsKey(job.Title);
            }
        }

        /// <summary>
        /// Returns true if status text or code has changed
        /// Returns true if new status or oldStatus is null
        /// </summary>
        /// <param name="newStatus">new statuses</param>
        /// <param name="oldStatus">old statuses</param>
        /// <returns>true if statuses has changed</returns>
        private bool StatusHasChanged(Status newStatus, Status oldStatus)
        {
            return (oldStatus == null) ||
                    (newStatus == null) ||
                    oldStatus.Changed(newStatus);
        }

        /// <summary>
        /// Register a job as running.
        /// Thread safe
        /// </summary>
        /// /// <param name="job">the job which run is being exectued</param>
        /// <param name="thread">the thread executing IJob.Run()</param>
        private void RegisterJobAsRunning(IJob job, Thread thread)
        {
            lock (workerThreads)
            {
                //workerThreads.Add(job.Title, thread);
                workerThreads.Add(job.Title, true);
            }
        }

        /// <summary>
        /// Unregister running job.
        /// Thread safe
        /// </summary>
        /// <param name="job">job</param>
        private void UnregisterJobAsRunning(IJob job)
        {
            lock (workerThreads)
            {
                if (workerThreads.ContainsKey(job.Title))
                    workerThreads.Remove(job.Title);
            }
        }

        /// <summary>
        /// Returns true if the job is running
        /// Thread safe
        /// </summary>
        /// <param name="job">job</param>
        /// <returns>bool : true if the job is running</returns>
        private bool JobIsRunning(IJob job)
        {
            lock (workerThreads)
            {
                return workerThreads.ContainsKey(job.Title);
            }
        }

        /// <summary>
        /// The method returns when all jobs have finished.
        /// No garantee that new jobs has not started after entering the method.
        /// This can potentially be an indefinate wait if new jobs are added while
        /// old ones still remain.
        /// </summary>
        public void WaitForAllStartedJobsToFinish()
        {


            int nrOfActiveJob;
            do
            {
                Thread.Sleep(0);
                lock (workerThreads)
                {
                    nrOfActiveJob = workerThreads.Count;
                }

            } while (nrOfActiveJob != 0);

            //Thread.Sleep(0);
            //Thread[] threadList;
            //lock(workerThreads)
            //{
            //    threadList = new Thread[workerThreads.Count];
            //    workerThreads.Values.CopyTo(threadList, 0);
            //}

            //foreach (Thread thread in threadList)
            //{
            //    if (thread.ManagedThreadId == Thread.CurrentThread.ManagedThreadId)
            //        continue;

            //    if (thread.IsAlive)
            //        thread.Join();
            //}
        }


        /// <summary>
        /// Starts the scheduler.
        /// The scheduler runs job at run-interval when on.
        /// </summary>
        public void Start()
        {
            if (!isRunning)
            {
                isRunning = true;
                RunJobs();
            }
        }

        /// <summary>
        /// Stops the scheduler.
        /// The scheduler does not run scheduled jobs when off.
        /// This does not prevent jobs from being run when
        /// RunJobs() is called.
        /// Stop only assures that upon completion of the method
        /// no new jobs will be started by the scheduler
        /// It does not garantee that jobs are not running upon completion.
        /// </summary>
        public void Stop()
        {
            isRunning = false;
            ClearAllTimers();
        }

        /// <summary>
        /// Remove all timers and hence stops automatic running of jobs.
        /// </summary>
        private void ClearAllTimers()
        {
            String[] titles = new string[timers.Count];
            timers.Keys.CopyTo(titles, 0);
            foreach (string title in titles)
            {
                timers[title].Dispose();
                timers.Remove(title);
            }
        }

        /// <summary>
        /// Should always be called before Scheduler is left for garbage collection.
        /// </summary>
        public void Close()
        {
            //By using JobTitles, a copy of the keys in the job dictionary
            //we insure that the removal of jobs won't effect the enumeration
            foreach (string title in JobTitles)
            {
                RemoveJob(title);
            }

        }

        /// <summary>
        /// True if the scheduler is running.
        /// Read-only property
        /// </summary>
        public bool IsRunning
        {
            get {return isRunning; }
        }

        /// <summary>
        /// Returns the statuses of a job given by its title
        /// Thread-safe
        /// </summary>
        /// <param name="title">the title of the job sought</param>
        /// <returns>statuses of the given job.</returns>
        public Status GetJobStatus(string title)
        {
            lock (statuses)
            {
                if (!statuses.ContainsKey(title))
                    throw new JobDoesNotExistException(title);

                return this.statuses[title];
            }
        }

        /// <summary>
        /// Sets the status of a job
        /// Thread-safe
        /// </summary>
        /// <param name="job">job</param>
        /// <param name="jobStatus">new statuses</param>
        private void SetJobStatus(IJob job, Status newStatus)
        {
            lock (statuses)
            {
                string key = job.Title;
                if (statuses.ContainsKey(key))
                    statuses[key] = newStatus;
            }
        }

        /// <summary>
        /// String-Array with titles of all jobs currently scheduled.
        /// </summary>
        public string[] JobTitles
        {
            get
            {
                string[] jobTitles = new string[jobs.Count];
                jobs.Keys.CopyTo(jobTitles, 0);
                return jobTitles;
            }
        }

        /// <summary>
        /// Array of all jobs running.
        /// Used for effiecent serialization
        /// </summary>
        public IJob[] Jobs
        {
            get
            {
                IJob[] jobList = new IJob[jobs.Count];
                jobs.Values.CopyTo(jobList, 0);
                return jobList;
            }
        }
    }
}
