﻿//**********************************//
// Written by: Per Böhlin
// perbo770@student.liu.se
// Copyright: (c) 2009
//
// Implementation started: July 11, 2009
// Last modified: July 23, 2009
//**********************************//

using System;
using System.Collections.Generic;
using System.Text;
using NUnit.Framework;
using DataBinding;
using DataBinding.Validators;
using DataBinding.Matchers;
using DataBindingTests.Mocks;

namespace DataBindingTests.TestCases
{
    [TestFixture]
    public class DataFieldTest
    {

        private bool eventTriggered;
        private string eventData;

        [Test]
        public void TestValidateInt()
        {
            const int expected = 3;
            
            DataField<int> testSubject = 
                new DataField<int>(new ValidatorMockValueNotZero<int>(), null);
            testSubject.Value = expected;

            Assert.AreEqual(expected, testSubject.Value);
            Assert.AreEqual(true, testSubject.IsValid());

            testSubject.Value = 0;
            Assert.AreEqual(false, testSubject.IsValid());
        }

        [Test]
        public void TestValidateDouble()
        {
            const double expected = 20.0;
            DataField<double> testSubject = 
                new DataField<double>(new ValidatorMockValueNotZero<double>(), null);
            testSubject.Value = expected;

            Assert.AreEqual(expected, testSubject.Value);
            Assert.AreEqual(true, testSubject.IsValid());

            testSubject.Value = 0.0;
            Assert.AreEqual(false, testSubject.IsValid());
       }

        [Test]
        public void TestValidateString()
        {
            const string expected = "message";
            
            DataField<string> testSubject = 
                new DataField<string>(new ValidatorMockValueNotNull<string>(), null);
            testSubject.Value = expected;

            Assert.AreEqual(expected, testSubject.Value);
            Assert.AreEqual(true, testSubject.IsValid());

            testSubject.Value = null;
            Assert.AreEqual(false, testSubject.IsValid());
        }

        private enum Values
        {
            ValidValue = 10,
            InvalidValue = 15
        }

        [Test]
        public void TestValidateEnum()
        {
            const Values expected = Values.ValidValue;
            DataField<Values> testSubject = 
                new DataField<Values>(new ExcludeValueValidator<Values>(Values.InvalidValue), null);
            testSubject.Value = expected;

            Assert.AreEqual(expected, testSubject.Value);
            Assert.AreEqual(true, testSubject.IsValid());

            testSubject.Value = Values.InvalidValue;
            Assert.AreEqual(false, testSubject.IsValid());

        }

        [Test]
        public void TestNullValidator()
        {
            const double expected = 20.0;
            DataField<double> testSubject = 
                new DataField<double>(null, null, expected);

            Assert.AreEqual(expected, testSubject.Value);
            Assert.AreEqual(true, testSubject.IsValid());

            testSubject.Value = 0.0;
            Assert.AreEqual(true, testSubject.IsValid());
        }

        [Test]
        public void TestNullMatcher()
        {
            const string expected = "text";
            DataField<string> testSubject = new DataField<string>(null, null, expected);
            DataField<string> searchField = new DataField<string>(null, null, expected);

            Assert.IsTrue(testSubject.IsMatch(searchField));
        }

        [Test]
        public void TestMatcher()
        {
            const string expected = "text";
            DataField<string> testSubject = 
                new DataField<string>(null, new EqualMatcher<string>(), expected);
            DataField<string> searchField = 
                new DataField<string>(null, null, expected);

            Assert.IsTrue(testSubject.IsMatch(searchField));

            searchField.Value = "other text";
            Assert.IsFalse(testSubject.IsMatch(searchField));
        }

        [Test]
        public void TestImplicitConvertionToDouble()
        {
            const double expected = 20.0;
            DataField<double> testSubject = 
                new DataField<double>(new ValidatorMockValueNotZero<double>(), null);
            testSubject.Value = expected;

            double retrievedValue = testSubject;
            Assert.AreEqual(expected, retrievedValue);
        }

        [Test]
        public void TestImplicitConvertionToString()
        {
            const string expected = "message";
            
            DataField<string> testSubject = 
                new DataField<string>(new ValidatorMockValueNotNull<string>(), null);
            testSubject.Value = expected;

            string retrievedValue = testSubject;
            Assert.AreEqual(expected, retrievedValue);
        }

        [Test]
        public void TestImplicitConvertionToEnum()
        {
            const Values expected = Values.ValidValue;

            DataField<Values> testSubject = 
                new DataField<Values>(new ExcludeValueValidator<Values>(Values.InvalidValue), null);
            testSubject.Value = expected;

            Values retrievedValue = testSubject;
            Assert.AreEqual(expected, retrievedValue);
        }

        [Test]
        public void TestCopyConstructorDouble()
        {
            const double expected = 20.0;
            DataField<double> target =
                new DataField<double>(new ValidatorMockValueNotZero<double>(), null);
            DataField<double> source =
                new DataField<double>(new ValidatorMockValueIsZero<double>(), null, expected);

            // assert failes validation on zero
            target.Value = 0;
            Assert.AreEqual(false, target.IsValid());
            target.Value = 1;
            Assert.AreEqual(true, target.IsValid());

            // create using copy constructor
            target = new DataField<double>(source);
            // asssert that target has new value
            Assert.AreEqual(expected, target.Value);
            // assert that target has new validator
            Assert.AreEqual(false, target.IsValid());
            // assert that target and source are not the same
            Assert.AreNotSame(target, source);
        }

        [Test]
        public void TestCopyConstructorString()
        {
            const string expected = "Text";
            DataField<string> target =  
                new DataField<string>(new ValidatorMockValueNotNull<string>(), null);
            DataField<string> source = 
                new DataField<string>(new ValidatorMockValueIsNull<string>(), null, expected);
           
            // assert failes validatation on null
            Assert.AreEqual(false, target.IsValid());
            target.Value = "";
            Assert.AreEqual(true, target.IsValid());
            
            // assert source behavior
            Assert.AreEqual(false, source.IsValid());

            // create using copy constructor
            target = new DataField<string>(source);
            // assert equal value
            Assert.AreEqual(expected, target.Value);
            // assert that target has new validator
            Assert.AreEqual(false, target.IsValid());
            // assert that target and source are not the same
            Assert.AreNotSame(target, source);
        }

        [Test]
        public void TestEqual()
        {
            // Two fields, same value, different validators
            DataField<string> source = 
                new DataField<string>(new Mocks.ValidatorMock<string>(""), null,"Alpha");
            DataField<string> target =
                new DataField<string>(new Mocks.ValidatorMockValueNotNull<string>(), null, "Alpha");
            
            // assert that referens do not point to the same instance
            Assert.AreNotSame(source, target);

            // assert equal in value
            Assert.AreEqual(source.Value, target.Value);

            // Method being tested
            Assert.IsTrue(source.Equals(target));
            
            // change value and assert dif.
            source.Value = "Bravo";
            Assert.AreNotEqual(source.Value, target.Value);

            // method being tested
            Assert.IsFalse(source.Equals(target));

            
        }

        [Test]
        public void TestEvents()
        {
            DataField<string> testSubject = new DataField<string>(null, null, "");
            
            //register eventhandler
            testSubject.Changed += this.OnChangeEvent;

            // init state
            eventTriggered = false;
            eventData = "";
            testSubject.Value = "text";
            Assert.IsTrue(eventTriggered);
            Assert.AreEqual(testSubject.Value, eventData);
        }

        private void OnChangeEvent(object sender, DataChangedEventArgs<string> args)
        {
            eventTriggered = true;
            eventData = args.NewValue;
        }
    }
}
