﻿//**********************************//
// Written by: Per Böhlin
// perbo770@student.liu.se
// Copyright: (c) 2009
//
// Implementation started: July 11, 2009
// Last modified: July 23, 2009
//**********************************//

using System;
using System.Collections.Generic;
using System.Text;
using DataBinding.Validators;
using DataBinding.Matchers;

namespace DataBinding
{
    /// <summary>
    /// Event handler delegate for DataFieldChanged event handlers
    /// </summary>
    /// <typeparam name="T">Should be the same type as the data field</typeparam>
    public delegate void DataFieldChangedEventHandler<T>(object sender, DataChangedEventArgs<T> args)
        where T: IComparable;
    
    /// <summary>
    /// Core class in the DataBinding framework.
    /// Triggers DataFieldChanged events
    /// Allows for implicit convertion TO its value type
    /// </summary>
    public class DataField<T> : IDataField<T> where T : IComparable
    {
        private T value;
        private readonly IValidator<T> validator;
        private readonly IMatcher<T> matcher;

        #region Constructors

        /// <summary>
        /// Default constructor
        /// </summary>
        public DataField()
            : this(default(T)){}

        public DataField(T value)
            : this(null, null, value){}

        public DataField(IValidator<T> validator, IMatcher<T> matcher)
            : this(validator, matcher, default(T)){}

        public DataField(IValidator<T> validator, IMatcher<T> matcher, T value)
        {
            this.validator = validator;
            this.matcher = matcher;
            this.value = value;
        }

        public DataField(DataField<T> otherField)
        {
            this.Value = otherField.Value;
            this.validator = otherField.validator;
            this.matcher = otherField.matcher;
        }

        #endregion

        #region Events
        public event DataFieldChangedEventHandler<T> Changed;

        protected virtual void OnChanged(DataChangedEventArgs<T> args)
        {
            if (Changed != null)
            {
                Changed(this, args);
            }
        }

        #endregion

        #region Properties
        public T Value
        {
            get { return value; }
            set {

                bool newValue = true;
                if (this.value != null && (this.value.Equals(value)))
                    newValue = false;

                this.value = value;
                if (newValue)
                    OnChanged(new DataChangedEventArgs<T>(this.value));
            }
        }
        #endregion

        #region Methods

        public bool IsValid()
        {
            return (validator != null) ? validator.IsValid(this.value) : true;
        }

        virtual public bool IsMatch(IDataField<T> searchField)
        {
            return (matcher != null) ? matcher.IsMatch(this.Value, searchField.Value) : true;
        }

        override public string ToString()
        {
            return (value != null) ? 
                value.ToString() :
                String.Empty;
        }

        override public bool Equals(object obj)
        {
            DataField<T> asField = obj as DataField<T>;

            if (asField != null)
            {
                return this.Equals(asField);
            }
            else
            {
                return false;
            }
        }

        public bool Equals(DataField<T> otherField)
        {
            if (otherField != null)
            {
                return this.value.Equals(otherField.value);
            }
            else
            {
                return false;
            }
        }

        public override int GetHashCode()
        {
            return base.GetHashCode() ^
                this.value.GetHashCode();
        }

        #endregion

        public static implicit operator T(DataField<T> dataField)
        {
            return dataField.Value;
        }

        

    }
}
