﻿//**********************************//
// Written by: Per Böhlin
// perbo770@student.liu.se
// Copyright: (c) 2009
//
// Implementation started: July 11, 2009
// Last modified: July 23, 2009
//**********************************//

using System;
using System.Collections.Generic;
using System.Text;

namespace DataBinding.FormControlValidators
{
    /// <summary>
    /// Abstact base class for FormControlValidators
    /// </summary>
    public abstract class FormControlValidatorBase<T> : IFormControlValidator where T: IComparable
    {
        protected IDataField<T> dataField;
        protected string invalidValueMessage;
        protected bool valid;
        
        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="dataField">Data field</param>
        /// <param name="invalidValueMessage"> message to be displayed if explcit validation fails</param>
        public FormControlValidatorBase(IDataField<T> dataField, string invalidValueMessage)
        {
            this.dataField = dataField;
            this.invalidValueMessage = invalidValueMessage;
            this.valid = false;
            this.dataField.Changed += this.DataField_ValueChanged;
        }

        /// <summary>
        /// Abstract method.
        /// Should return true if value parsing between control and field succeeded and
        /// value is valid.
        /// Subclass implementation should call UpdateAndValidateValue(T value)
        /// </summary>
        /// <returns></returns>
        public abstract bool Validate();

        /// <summary>
        /// Returns error mesage if an error has occured.
        /// Otherwise empty string
        /// </summary>
        /// <returns>error message or empty string</returns>
        public string GetLastErrorMessage()
        {
            if (this.valid)
                return string.Empty;

            this.valid = true;
            return invalidValueMessage;
        }

        /// <summary>
        /// Should be called from subclass.
        /// Updates datafield with value and causes a validation.
        /// </summary>
        /// <param name="value">new data field value</param>
        protected void UpdateAndValidateValue(T value)
        {
            dataField.Value = value;
            ValidateValue();
        }

        /// <summary>
        /// Should be called by UpdateAndValidateValue and any overriding versions
        /// Sets internal state variable valid.
        /// </summary>
        protected void ValidateValue()
        {
            this.valid &= dataField.IsValid();
        }

        /// <summary>
        /// Event handler: DataField's value changed
        /// Updates the controls value when data fields value change
        /// </summary>
        virtual protected void DataField_ValueChanged(object sender, DataChangedEventArgs<T> e)
        {
            this.ReverseUpdateAndValidateValue();
        }

        /// <summary>
        /// Abstract method.
        /// Update procedure for how data fields value
        /// can update the value of the form control
        /// </summary>
        abstract protected void ReverseUpdateAndValidateValue();

        virtual public void Dispose()
        {
            this.dataField.Changed -= this.DataField_ValueChanged;
        }
    }
}
