﻿//**********************************//
// Written by: Per Böhlin
// perbo770@student.liu.se
// Copyright: (c) 2009
//
// Implementation started: July 27, 2009
// Last modified: August 12, 2009
//**********************************//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Net;
using DataBinding;
using DataBinding.FormControlValidators;
using DataBinding.Validators;

namespace Hudson
{
    /// <summary>
    /// Form that lets the user configure a Hudson Job.
    /// </summary>
    partial class ConfigFrame : Form
    {
        private List<IFormControlValidator> validators;
        private List<IFormControlValidator> credentialValidators;

        DataField<string> urlFeed;
        DataField<string> urlLaunch;
        DataField<int> timeout;
        DataField<HudsonSettings.AuthenticationMethods> authenticationMethod;
        DataField<string> userName;
        DataField<string> password;
        DataField<string> filterPattern;

        HudsonSettings settings;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="title">Title of the job being configured</param>
        /// <param name="settings">Hudson settings object, current settings</param>
        public ConfigFrame(string title, HudsonSettings settings, FormStartPosition startPosition)
        {
            // GUI Designer initilizations
            InitializeComponent();

            //Custom initilizations
            InitializeGUI(title, startPosition);

            //Initilize fields
            InitializeFields(settings);

            DataBinding();

            UpdateForm();
        }

        /// <summary>
        /// Initialize GUI
        /// </summary>
        /// <param name="title">title of the job being configured</param>
        private void InitializeGUI(string title, FormStartPosition startPosition)
        {
            this.StartPosition = startPosition;
            this.ShowInTaskbar = (startPosition == FormStartPosition.CenterScreen);
            this.Text = Properties.Resources.ConfigFormCaption + ": " + title;
            this.Icon = Properties.Resources.Hudson_icon;
            this.FormClosing += ConfigFrame_FormClosing;
            txtUrlFeed.Text = "";
            txtLaunch.Text = "";
            txtConnectionTimeout.Text = "";
            txtUserName.Text = "";
            txtPassword.Text = "";
            txtFilter.Text = "";
            cmbAuthenticationMethod.DataSource = Enum.GetNames(typeof(HudsonSettings.AuthenticationMethods));
        }

        /// <summary>
        /// Initialize Fields
        /// </summary>
        /// <param name="settings">Hudson settings object</param>
        private void InitializeFields(HudsonSettings settings)
        {
            this.settings = new HudsonSettings(settings);

            this.validators = new List<IFormControlValidator>();
            this.credentialValidators = new List<IFormControlValidator>();

            this.urlFeed = new DataField<string>(new NoneEmptyStringValidator(),
                null, settings.UrlFeed);

            this.urlLaunch = new DataField<string>(null,
                null, settings.Launch);

            this.timeout = new DataField<int>(new NoneNegativeValidator<int>(), null, settings.Timeout);
            this.authenticationMethod = new DataField<HudsonSettings.AuthenticationMethods>(null, null, settings.AuthenticationMethod);
            this.userName = new DataField<string>(new NoneEmptyStringValidator(),
                null, settings.UserName);
            this.password = new DataField<string>(new NoneEmptyStringValidator(),
                null, settings.Password);
            this.filterPattern = new DataField<string>(new IsRegexValidator(), null, settings.FilterPattern);
        }

        /// <summary>
        /// Bind form controls to datafields for validation
        /// </summary>
        private void DataBinding()
        {
            validators.Clear();
            credentialValidators.Clear();

            IFormControlValidator validator;

            validator = new TextBoxValidatorColor<string>(this.txtUrlFeed, this.urlFeed,
                Properties.Resources.InvalidInputUrlFeedMessage, Color.Red);
            validators.Add(validator);

            validator = new TextBoxValidatorColor<string>(this.txtLaunch, this.urlLaunch,
                Properties.Resources.InvalidInputUrlLaunchMessage, Color.Red);
            validators.Add(validator);

            validator = new TextBoxValidatorColor<int>(this.txtConnectionTimeout,
                this.timeout, Properties.Resources.InvalidInputTimeoutMessage, Color.Red);
            validators.Add(validator);

            validator = new ComboBoxIndexValidatorColor<HudsonSettings.AuthenticationMethods>(this.cmbAuthenticationMethod,
                this.authenticationMethod, Properties.Resources.InvalidInputAuthenticationMethodMessage, Color.Red);
            validators.Add(validator);

            validator = new TextBoxValidatorColor<string>(this.txtUserName, this.userName,
                Properties.Resources.InvalidInputUserNameMessage, Color.Red);
            credentialValidators.Add(validator);

            validator = new TextBoxValidatorColor<string>(this.txtPassword, this.password,
                Properties.Resources.InvalidInputPasswordMessage, Color.Red);
            credentialValidators.Add(validator);

            validator = new TextBoxValidatorColor<string>(this.txtFilter, this.filterPattern,
                Properties.Resources.InvalidInputFilterPatternMessage, Color.Red);
            validators.Add(validator);
        }

        /// <summary>
        /// Update GUI
        /// </summary>
        private void UpdateForm()
        {
            bool credRequired = CredentialsRequired();
            txtUserName.Enabled = credRequired;
            txtPassword.Enabled = credRequired;
            lblPasswordWarning.Enabled = credRequired;
            if (!credRequired)
            {
                txtUserName.Text = "";
                txtPassword.Text = "";
            }
        }

        /// <summary>
        /// Returns true if user has selected an authentication method.
        /// </summary>
        /// <returns>true if authentication method has been selected to other than None</returns>
        private bool CredentialsRequired()
        {
            return (this.cmbAuthenticationMethod.SelectedIndex != (int) HudsonSettings.AuthenticationMethods.None);
        }

        /// <summary>
        /// Eventhandler for when Authentication Schema is changed.
        /// </summary>
        private void cmbAuthenticationSchema_SelectedIndexChanged(object sender, EventArgs e)
        {
            UpdateForm();
        }

        /// <summary>
        /// Settings object representing the configuration selected by the user
        /// Read only property
        /// </summary>
        public HudsonSettings Settings
        {
            get
            {
                if (this.DialogResult == DialogResult.OK)
                {
                    settings.UrlFeed = this.urlFeed;
                    settings.Launch = this.urlLaunch;
                    settings.Timeout = this.timeout;
                    settings.AuthenticationMethod = this.authenticationMethod;
                    settings.UserName = this.userName;
                    settings.Password = this.password;
                    settings.FilterPattern = this.filterPattern;
                }
                return this.settings;
            }
        }

        /// <summary>
        /// Validates all FormControlValidators
        /// Shows error message if any would fail
        /// </summary>
        /// <returns>Returns true if all validators validates, otherwise false</returns>
        private bool ValidateAll()
        {
            foreach (IFormControlValidator validator in validators)
            {
                if (!validator.Validate())
                {
                    MessageBox.Show(validator.GetLastErrorMessage(), Properties.Resources.InvalidInputCaption);
                    return false;
                }
            }
            if (CredentialsRequired())
            {
                foreach (IFormControlValidator validator in credentialValidators)
                {
                    if (!validator.Validate())
                    {
                        MessageBox.Show(validator.GetLastErrorMessage(), Properties.Resources.InvalidInputCaption);
                        return false;
                    }
                }
            }

            return true;
        }

        /// <summary>
        /// Eventhandler for when the form closes.
        /// Validates all FormControlValidators.
        /// Cancels the closing event if some thing don't validate.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ConfigFrame_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (this.DialogResult == DialogResult.OK)
                e.Cancel = !ValidateAll();
        }

    }
}