﻿//**********************************//
// Written by: Per Böhlin
// perbo770@student.liu.se
// Copyright: (c) 2009
//
// Implementation started: July 27, 2009
// Last modified: August 12, 2009
//**********************************//

using System;
using System.Collections.Generic;
using System.Text;
using BSM.core;
using System.ServiceModel.Syndication;
using System.Xml;
using System.Xml.Serialization;
using System.IO;
using System.Drawing;
using System.Net;
using System.Text.RegularExpressions;
using System.Windows.Forms;

namespace Hudson
{
    /// <summary>
    /// Hudson Job
    /// Parses a Rss/Atom feed to return build status from a Hudson Build System.
    /// Hudson - https://hudson.dev.java.net/
    /// </summary>
    public class HudsonJob : JobBase
    {
        /// <summary>
        /// Used to describe the status of a read RSS/Atom feed
        /// </summary>
        [Flags]
        private enum FeedStatus
        {
            InvalidFeed = 0,
            FoundFeedItems = 1,
            FoundFilteredFeedItems = 2
        }

        private HudsonSettings settings;

        /// <summary>
        /// Default constructor
        /// </summary>
        public HudsonJob()
            : this(new HudsonSettings())
        {
        }

        /// <summary>
        /// Load constructor
        /// </summary>
        /// <param name="settings">Hudson settings object</param>
        public HudsonJob(HudsonSettings settings)
            : base("Hudson")
        {
            this.settings = settings;
        }

        /// <summary>
        /// Configure the job
        /// Requires user interaction
        /// </summary>
        /// <param name="owner">owner</param>
        /// <returns>true if changes has been made</returns>
        override public bool Configure(IWin32Window owner)
        {
            FormStartPosition startPosition = (owner != null) ?
                FormStartPosition.CenterParent :
                FormStartPosition.CenterScreen;

            ConfigFrame form = new ConfigFrame(this.Title, settings, startPosition);
            if (form.ShowDialog(owner) == System.Windows.Forms.DialogResult.OK)
            {
                settings = form.Settings;
                return true;
            }
            return false;
        }

        /// <summary>
        /// Load job settings from file, using file path
        /// Throws:
        ///     SettingsException
        /// </summary>
        /// <param name="filePath">path to settings file</param>
        override public void LoadSettings(string filePath)
        {
            try
            {
                XmlSerializer serializer = new XmlSerializer(typeof(HudsonSettings));
                using (TextReader reader = new StreamReader(filePath))
                {
                    this.settings = (HudsonSettings)serializer.Deserialize(reader);
                    reader.Close();
                }
            }
            catch (Exception ex)
            {
                throw new SettingsException(Properties.Resources.SettingsLoadErrorMessage, ex);
            }
        }

        /// <summary>
        /// Save settings to file, using file path
        /// Throws:
        ///     SettingsException
        /// </summary>
        /// <param name="filePath">path to settings file</param>
        override public void SaveSettings(string filePath)
        {
            try
            {
                XmlSerializer serializer = new XmlSerializer(typeof(HudsonSettings));
                using (TextWriter writer = new StreamWriter(filePath))
                {
                    serializer.Serialize(writer, this.settings);
                    writer.Close();
                }
            }
            catch (Exception ex)
            {
                throw new SettingsException(Properties.Resources.SettingsSaveErrorMessage, ex);
            }
        }

        /// <summary>
        /// Deletes the jobs settings file, given its file path
        /// Throws SettingsException
        /// </summary>
        /// <param name="filePath">file path to settings file</param>
        override public void DeleteSettings(string filePath)
        {
            if (!File.Exists(filePath))
                throw new SettingsException(String.Format(Properties.Resources.SettingsFileNotFoundErrorFormat, filePath), new FileNotFoundException());

            File.Delete(filePath);
        }

        /// <summary>
        /// Run Job
        /// Reads the RSS/Atom feed provided through settings
        /// Parse it and returns a Status object.
        /// Throws:
        ///     RunFailedException
        /// </summary>
        /// <returns></returns>
        override public Status Run()
        {
            try
            {
                string title = String.Empty;
                FeedStatus feedStatus = FeedStatus.InvalidFeed;
                DateTimeOffset lastPubDate = default(DateTimeOffset);

                HttpWebRequest request = (HttpWebRequest)WebRequest.Create(settings.UrlFeed);

                if (CredentialsRequired(settings))
                {
                    request.Credentials = GetCredentials(settings);
                }

                request.Timeout = settings.Timeout;
                using (HttpWebResponse response = (HttpWebResponse)request.GetResponse())
                {
                    using (XmlReader reader = XmlReader.Create(response.GetResponseStream()))
                    {
                        ReadFeed(ref title, ref feedStatus, ref lastPubDate, reader);
                        reader.Close();
                    }
                }
                Status result = GetRunResult(title, feedStatus);

                return result;
            }
            catch (TimeoutException ex)
            {
                throw new RunFailedException(Properties.Resources.RunFailedTimeoutMessage, ex);
            }
            catch (System.Net.WebException ex)
            {
                RunFailedException wrappedException = WebExceptionHandling(ex);
                throw wrappedException;
            }
            catch (Exception ex)
            {
                throw new RunFailedException(Properties.Resources.RunFailedUnknownMessage, ex);
            }
        }

        /// <summary>
        /// Wraps a WebException in a RunFailedException with custom message
        /// depending on type of web error.
        /// </summary>
        /// <param name="ex">WebException</param>
        /// <returns>RunFailedException with custom error message</returns>
        private RunFailedException WebExceptionHandling(WebException ex)
        {
            if (ex.Message.Contains("401"))
            {
                return new RunFailedException(Properties.Resources.RunFailedAuthenticationRequiredMessage, ex);
            }
            if (ex.Message.Contains("404"))
            {
                return new RunFailedException(Properties.Resources.RunFailedNotFoundMessage, ex);
            }
            string message = Properties.Resources.RunFailedNetworkErrorMessage;
            message += " " + ex.Message;
            return new RunFailedException(message, ex);
        }

        /// <summary>
        /// Reads RSS/Atom feed from reader : XmlReader
        /// </summary>
        /// <param name="title">Title of newest atom item</param>
        /// <param name="feedStatus">status of the feed</param>
        /// <param name="lastPubDate">pubData of newest atom item</param>
        /// <param name="reader">reader read from</param>
        private void ReadFeed(ref string title, ref FeedStatus feedStatus, ref DateTimeOffset lastPubDate, XmlReader reader)
        {
            SyndicationFeed feed = SyndicationFeed.Load(reader);
            Regex filter = new Regex(settings.FilterPattern);

            foreach (SyndicationItem item in feed.Items)
            {
                if (lastPubDate < item.PublishDate)
                {
                    feedStatus |= FeedStatus.FoundFeedItems;
                    if (filter.IsMatch(item.Title.Text))
                    {
                        lastPubDate = item.PublishDate;
                        title = item.Title.Text;
                        feedStatus |= FeedStatus.FoundFilteredFeedItems;
                    }
                }
            }
        }

        /// <summary>
        /// Get credentials for web authentication
        /// </summary>
        /// <param name="request">Credentials</param>
        private CredentialCache GetCredentials(HudsonSettings settings)
        {
            NetworkCredential credential = new NetworkCredential(settings.UserName, settings.Password);
            CredentialCache cache = new CredentialCache();
            cache.Add(new Uri(settings.UrlFeed), this.settings.AuthenticationMethod.ToString(), credential);
            return cache;
        }

        /// <summary>
        /// Returns true if credentials are required based on settings
        /// </summary>
        /// <returns>true if credentials are required</returns>
        private bool CredentialsRequired(HudsonSettings settings)
        {
            return settings.AuthenticationMethod != HudsonSettings.AuthenticationMethods.None;
        }

        /// <summary>
        /// Interprets Atom item title and Feed status
        /// and returns a Status object accordingly
        /// </summary>
        /// <param name="title">title of an Atom item : string</param>
        /// <param name="feedStatus">Status of Rss/Atom feed parsed</param>
        /// <returns></returns>
        private Status GetRunResult(string title, FeedStatus feedStatus)
        {
            Status result = new Status();

            // Unstable first since unstable matches stable's token as well
            if (title.Contains(Properties.Settings.Default.DefaultUnstableToken))
            {
                result.Code = Status.StatusCode.Yellow;
                result.Text = title;
            }
            else if (title.Contains(Properties.Settings.Default.DefaultSuccessToken) ||
                title.Contains(Properties.Settings.Default.DefaultNormalToken))
            {
                result.Code = Status.StatusCode.Green;
                result.Text = title;
            }
            else if (title.Contains(Properties.Settings.Default.DefaultFailToken) ||
                     title.Contains(Properties.Settings.Default.DefaultBrokenToken))
            {
                result.Code = Status.StatusCode.Red;
                result.Text = title;
            }
            else if ((feedStatus & FeedStatus.FoundFilteredFeedItems) == FeedStatus.FoundFilteredFeedItems)
            {
                result.Code = Status.StatusCode.Unknown;
                result.Text = title;
            }
            else if ((feedStatus & FeedStatus.FoundFeedItems) == FeedStatus.FoundFeedItems)
            {
                result.Code = Status.StatusCode.Unknown;
                result.Text = Properties.Resources.StatusTextNoFilteredItemsFound;
            }
            else
            {
                result.Code = Status.StatusCode.Unknown;
                result.Text = Properties.Resources.StatusTextNotAtomFeed;
            }
            
            return result;
        }

        /// <summary>
        /// Launch application provided in job configuration
        /// </summary>
        public override void Launch()
        {
            try
            {
                if (settings.Launch != String.Empty)
                    System.Diagnostics.Process.Start(settings.Launch);
            }
            catch
            {
                MessageBox.Show(String.Format(Properties.Resources.LaunchErrorFormat, settings.Launch),
                    Properties.Resources.ErrorMessageCaption);
            }
        }
    }
}
